"""Initialize testing module, specify used module versions."""
import gi

gi.require_version("Adw", "1")
gi.require_version("Gdk", "4.0")
gi.require_version("Graphs", "1")
gi.require_version("Gtk", "4.0")



# SPDX-License-Identifier: GPL-3.0-or-later
"""Miscallaneous constants."""

from gettext import pgettext as C_

from gi.repository import Graphs

# Type hints
ItemList = list[Graphs.Item]
Limits = tuple[float, float, float, float]


class ParseError(Exception):
    """Custom Error for parsing files."""

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


LEGEND_POSITIONS = [
    "best",
    "upper right",
    "upper left",
    "lower left",
    "lower right",
    "center left",
    "center right",
    "lower center",
    "upper center",
    "center",
]
DELIMITERS = {
    "whitespace": "\\s+",
    "tabs": "\\t+",
    "colon": ":",
    "semicolon": ";",
    "comma": ",",
    "period": ".",
    "custom": "custom",
}
LINESTYLES = ["none", "solid", "dotted", "dashed", "dashdot"]
MARKERSTYLES = [
    "none",
    ".",
    ",",
    "o",
    "v",
    "^",
    "<",
    ">",
    "8",
    "s",
    "p",
    "*",
    "h",
    "H",
    "+",
    "x",
    "D",
    "d",
    "|",
    "_",
    "P",
    "X",
]
LIMITS = [
    "min_bottom",
    "max_bottom",
    "min_top",
    "max_top",
    "min_left",
    "max_left",
    "min_right",
    "max_right",
]

DIRECTIONS = ["bottom", "top", "left", "right"]

FUNCTIONS = {
    "sin",
    "cos",
    "tan",
    "sinh",
    "cosh",
    "tanh",
    "arcsin",
    "arccos",
    "arctan",
    "arcsinh",
    "arccosh",
    "arctanh",
    "cot",
    "sec",
    "csc",
    "arccot",
    "arcsec",
    "arccsc",
    "sqrt",
    "exp",
    "abs",
    "log",
    "log10",
}

GRAPHS_PROJECT_FILE_FILTER_TEMPLATE = \
    (C_("file-filter", "Graphs Project File"), ["graphs"])



# SPDX-License-Identifier: GPL-3.0-or-later
"""Main application."""
import logging
from gettext import gettext as _

from gi.repository import Adw, Gio, Graphs, Gtk

from graphs import (
    actions,
    file_import,
    migrate,
    operations,
    project,
    styles,
    ui,
)
from graphs.data import Data

from matplotlib import font_manager


class PythonApplication(Graphs.Application):
    """The main application singleton class."""

    __gtype_name__ = "GraphsPythonApplication"

    def __init__(self, application_id, **kwargs):
        settings = Gio.Settings(application_id)
        migrate.migrate_config(settings)
        data = Data(settings.get_child("figure"))
        super().__init__(
            application_id=application_id,
            settings=settings,
            flags=Gio.ApplicationFlags.HANDLES_OPEN,
            data=data,
            **kwargs,
        )
        font_list = font_manager.findSystemFonts(fontpaths=None, fontext="ttf")
        for font in font_list:
            try:
                font_manager.fontManager.addfont(font)
            except RuntimeError:
                logging.warning(_("Could not load {font}").format(font=font))

        Graphs.setup_actions(self)
        self.connect("action_invoked", actions.on_action_invoked)
        self.connect("operation_invoked", operations.perform_operation)

        data.connect(
            "notify::items",
            ui.on_items_change,
            self,
        )

    def on_project_saved(self, _application, handler=None, *args) -> None:
        """Change unsaved state."""
        self.disconnect(self.save_handler)
        if handler == "close":
            self.quit()
        data = self.get_data()
        if handler == "open_project":
            data.props.project_file = args[0]
            data.load()

    def do_open(self, files: list, nfiles: int, _hint: str) -> None:
        """Open Graphs with a File as argument."""
        self.do_activate()
        data = self.get_data()
        if nfiles == 1 and files[0].get_uri().endswith(".graphs"):
            project_file = files[0]
            if data.get_unsaved():

                def on_response(_dialog, response):
                    if response == "discard_close":
                        project.load(data, project_file)
                    if response == "save_close":
                        self.save_handler = data.connect(
                            "saved",
                            self.on_project_saved,
                            "open_project",
                            project_file,
                        )
                        project.save_project(self)

                dialog = Graphs.tools_build_dialog("save_changes")
                dialog.set_transient_for(self.get_window())
                dialog.connect("response", on_response)
                dialog.present()

            else:
                project.load(data, project_file)
        else:
            file_import.import_from_files(self, files)

    def close_application(self, *_args) -> None:
        """
        Intercept when closing the application.

        Will ask the user to confirm
        and save/discard open data if any unsaved changes are present.
        """
        if self.get_data().get_unsaved():

            def on_response(_dialog, response):
                if response == "discard_close":
                    self.quit()
                if response == "save_close":
                    self.save_handler = self.get_data().connect(
                        "saved",
                        self.on_project_saved,
                        "close",
                    )
                    project.save_project(self)

            dialog = Graphs.tools_build_dialog("save_changes")
            dialog.connect("response", on_response)
            dialog.present(self.get_window())
            return True
        self.quit()

    def do_activate(self) -> None:
        """
        Activate the application.

        We raise the application"s main window, creating it if
        necessary.
        """
        window = self.props.active_window
        if not window:
            window = Graphs.Window.new(self)
            data = self.get_data()
            binding_table = [
                ("can_undo", window.get_undo_button(), "sensitive"),
                ("can_redo", window.get_redo_button(), "sensitive"),
                ("can_view_back", window.get_view_back_button(), "sensitive"),
                (
                    "can_view_forward",
                    window.get_view_forward_button(),
                    "sensitive",
                ),
                ("project_name", window.get_content_title(), "title"),
                ("project_path", window.get_content_title(), "subtitle"),
            ]
            for prop1, obj, prop2 in binding_table:
                data.bind_property(prop1, obj, prop2, 2)
            actions = (
                "multiply_x",
                "multiply_y",
                "translate_x",
                "translate_y",
            )
            for action in actions:
                entry = window.get_property(action + "_entry")
                button = window.get_property(action + "_button")
                entry.connect(
                    "notify::text",
                    self.set_entry_css,
                    entry,
                    button,
                )
                data.connect(
                    "notify::items-selected",
                    self.set_entry_css,
                    entry,
                    button,
                )
                self.set_entry_css(None, None, entry, button)
            self.set_window(window)
            window.connect("close-request", self.close_application)
            self.set_figure_style_manager(styles.StyleManager(self))
            data.connect_after(
                "notify::items",
                lambda _data,
                _param: window.update_view_menu(),
            )
            window.update_view_menu()
            window.present()

    def set_entry_css(
        self,
        _object,
        _param,
        entry: Adw.EntryRow,
        button: Gtk.Button,
    ) -> None:
        """Validate text field input."""
        button.set_sensitive(
            ui.validate_entry(entry)[1]
            and self.get_data().props.items_selected,
        )



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for parsing and writing styles."""
import logging
import typing
from gettext import gettext as _

from gi.repository import Gio

from matplotlib import RcParams, cbook, rc_context
from matplotlib.figure import Figure
from matplotlib.font_manager import font_scalings, weight_dict
from matplotlib.style.core import STYLE_BLACKLIST

import numpy


STYLE_IGNORELIST = [
    "savefig.dpi",
    "savefig.facecolor",
    "savefig.edgecolor",
    "savefig.format",
    "savefix.bbox",
    "savefig.pad_inches",
    "savefig.transparent",
    "savefig.orientation",
]
FONT_SIZE_KEYS = [
    "font.size",
    "axes.labelsize",
    "xtick.labelsize",
    "ytick.labelsize",
    "legend.fontsize",
    "figure.labelsize",
    "figure.titlesize",
    "axes.titlesize",
]


def parse(file: Gio.File) -> (RcParams, str):
    """
    Parse a style to RcParams.

    This is an improved version of matplotlibs '_rc_params_in_file()' function.
    It is also modified to work with GFile instead of the python builtin
    functions.
    """
    style = RcParams()
    filename = file.get_basename()
    try:
        stream = Gio.DataInputStream.new(file.read(None))
        line_number = 0
        while True:
            line = stream.read_line_utf8(None)[0]
            if line is None:
                break
            line_number += 1
            if line_number == 2:
                name = line[2:]
            line = cbook._strip_comment(line)
            if not line:
                continue
            try:
                key, value = line.split(":", 1)
            except ValueError:
                msg = _("Missing colon in file {file}, line {line}")
                logging.warning(msg.format(file=filename, line=line_number))
                continue
            key = key.strip()
            value = value.strip()
            if value.startswith('"') and value.endswith('"'):
                value = value[1:-1]  # strip double quotes
            if key in STYLE_BLACKLIST:
                msg = _("Non-style related parameter {param} in file {file}")
                logging.warning(msg.format(param=key, file=filename))
            elif key in STYLE_IGNORELIST:
                msg = _("Ignoring parameter {param} in file {file}")
                logging.warning(msg.format(param=key, file=filename))
            elif key in style:
                msg = _("Duplicate key in file {file}, on line {line}")
                logging.warning(msg.format(file=filename, line=line_number))
            else:
                if key in FONT_SIZE_KEYS \
                        and not value.replace(".", "", 1).isdigit():
                    try:
                        value = font_scalings[value]
                    except KeyError:
                        continue
                elif key == "font.weight" and not value.isdigit():
                    try:
                        value = weight_dict[value]
                    except KeyError:
                        continue
                try:
                    style[key] = value
                except (KeyError, ValueError):
                    msg = _("Bad value in file {file} on line {line}")
                    logging.exception(
                        msg.format(file=filename, line=line_number),
                    )
    except UnicodeDecodeError:
        logging.exception(
            _("Could not parse {filename}").format(filename=filename),
        )
    finally:
        stream.close()
    return style, name


WRITE_IGNORELIST = STYLE_IGNORELIST + [
    "lines.dashdot_pattern",
    "lines.dashed_pattern",
    "lines.dotted_pattern",
    "lines.dash_capstyle",
    "lines.dash_joinstyle",
    "lines.solid_capstyle",
    "lines.solid_joinstyle",
]


def write(file: Gio.File, name: str, style: RcParams) -> None:
    """Write a style to a file."""
    if file.query_exists(None):
        file.delete(None)
    stream = Gio.DataOutputStream.new(file.create(0, None))
    stream.put_string("# Generated via Graphs\n")
    stream.put_string(f"# {name}\n")
    for key, value in style.items():
        if key not in STYLE_BLACKLIST and key not in WRITE_IGNORELIST:
            value = str(value).replace("#", "")
            if key != "axes.prop_cycle":
                value = value.replace("[", "").replace("]", "")
                value = value.replace("'", "").replace("'", "")
                value = value.replace('"', "").replace('"', "")
            stream.put_string(f"{key}: {value}\n")
    stream.close()


_PREVIEW_XDATA = numpy.linspace(0, 10, 30)
_PREVIEW_YDATA1 = numpy.sin(_PREVIEW_XDATA)
_PREVIEW_YDATA2 = numpy.cos(_PREVIEW_XDATA)


def create_preview(
    file: typing.IO,
    params: RcParams,
    file_format: str = "svg",
) -> None:
    """Create preview of params and write it to file."""
    with rc_context(params):
        # set render size in inch
        figure = Figure(figsize=(5, 3))
        axis = figure.add_subplot()
        axis.spines.bottom.set_visible(True)
        axis.spines.left.set_visible(True)
        if not params["axes.spines.top"]:
            axis.tick_params(which="both", top=False, right=False)
        axis.plot(_PREVIEW_XDATA, _PREVIEW_YDATA1)
        axis.plot(_PREVIEW_XDATA, _PREVIEW_YDATA2)
        axis.set_xlabel(_("X Label"))
        axis.set_xlabel(_("Y Label"))
        figure.savefig(file, format=file_format)



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Module for importing data from files.

    Functions:
        import_from_files
"""
from pathlib import Path

from gi.repository import Gio, Graphs

from graphs import parse_file
from graphs.misc import ParseError

_IMPORT_MODES = {
    # name: suffix
    "project": ".graphs",
    "xrdml": ".xrdml",
    "xry": ".xry",
    "columns": None,
}


def import_from_files(
    application: Graphs.Application,
    files: list[Gio.File],
) -> None:
    """
    Import from a list of files.

    Automatically guesses, which mode to use. If configurable settings are
    present at /se/sjoerd/Graphs/import-params, a Window will be shown,
    giving the option to configure them.
    """
    import_dict = {mode: [] for mode in _IMPORT_MODES.keys()}
    for file in files:
        import_dict[_guess_import_mode(file)].append(file)
    modes = []
    for mode, files in import_dict.items():
        if files:
            modes.append(mode)
    configurable_modes = []
    settings = application.get_settings_child("import-params")
    for mode in settings.list_children():
        if mode in modes:
            configurable_modes.append(mode)
    if configurable_modes:

        def on_accept(_dialog):
            _import_from_files(
                application,
                settings,
                configurable_modes,
                import_dict,
            )

        dialog = Graphs.ImportDialog.new(application, configurable_modes)
        dialog.connect("accept", on_accept)
    else:
        _import_from_files(
            application,
            settings,
            configurable_modes,
            import_dict,
        )


def _import_from_files(
    application: Graphs.Application,
    settings: Gio.Settings,
    configurable_modes: list[str],
    import_dict: dict,
):
    items = []
    style = application.get_figure_style_manager().get_selected_style_params()
    for mode, files in import_dict.items():
        callback = getattr(parse_file, "import_from_" + mode)
        params = settings.get_child(mode) if mode in configurable_modes \
            else None
        for file in files:
            try:
                items.extend(callback(params, style, file))
            except ParseError as error:
                application.get_window().add_toast_string(error.message)
                continue
    application.get_data().add_items(
        items,
        application.get_figure_style_manager(),
    )


def _guess_import_mode(file: Gio.File) -> str:
    try:
        filename = Graphs.tools_get_filename(file)
        file_suffix = Path(filename).suffixes[-1]
    except IndexError:
        file_suffix = None
    for mode, suffix in _IMPORT_MODES.items():
        if suffix is not None and file_suffix == suffix:
            return mode
    return "columns"



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for data transformations."""
import logging
import sys
from gettext import gettext as _

from gi.repository import Graphs

from graphs import misc, utilities
from graphs.item import DataItem

import numexpr

import numpy

import scipy


def get_data(application: Graphs.Application, item: DataItem):
    """
    Retrieve item from datadict with start and stop index.

    If interaction_mode is set to "SELECT"
    """
    new_xdata = item.props.xdata
    new_ydata = item.props.ydata

    canvas = application.get_window().get_canvas()
    if canvas.get_mode() == 2:
        figure_settings = application.get_data().get_figure_settings()
        if item.get_xposition() == 0:
            xmin = figure_settings.get_min_bottom()
            xmax = figure_settings.get_max_bottom()
            scale = figure_settings.get_bottom_scale()
        else:
            xmin = figure_settings.get_min_top()
            xmax = figure_settings.get_max_top()
            scale = figure_settings.get_top_scale()
        startx = utilities.get_value_at_fraction(
            figure_settings.get_min_selected(),
            xmin,
            xmax,
            scale,
        )
        stopx = utilities.get_value_at_fraction(
            figure_settings.get_max_selected(),
            xmin,
            xmax,
            scale,
        )
        # If startx and stopx are not out of range, that is,
        # if the item data is within the highlight
        new_min = min(new_xdata)
        a = startx < new_min and stopx < new_min
        if not (a or (startx > max(new_xdata))):
            new_xdata, new_ydata = filter_data(
                new_xdata, new_ydata, ">=", startx,
            )
            new_xdata, new_ydata = filter_data(
                new_xdata, new_ydata, "<=", stopx,
            )
        else:
            new_xdata = None
            new_ydata = None
    return new_xdata, new_ydata


def filter_data(
    xdata: list,
    ydata: list,
    condition: str,
    value: float,
) -> list:
    """Filter coordinates based on the given condition."""
    xdata = numpy.array(xdata)
    ydata = numpy.array(ydata)

    conditions = {
        "<=": numpy.less_equal,
        ">=": numpy.greater_equal,
        "==": numpy.equal,
    }

    mask = conditions[condition](xdata, value)

    xdata_filtered = xdata[mask]
    ydata_filtered = ydata[mask]

    return list(xdata_filtered), list(ydata_filtered)


def create_data_mask(xdata1: list, ydata1: list, xdata2: list, ydata2: list):
    """
    Create a mask for matching pairs of coordinates.

    Returns:
    - Boolean mask indicating where pairs of coordinates match.
    """
    xdata1, ydata1, xdata2, ydata2 = \
        map(numpy.array, [xdata1, ydata1, xdata2, ydata2])
    return numpy.any((xdata1[:, None] == xdata2) & (ydata1[:, None] == ydata2),
                     axis=1)


def sort_data(xdata: list, ydata: list) -> (list, list):
    """Sort data."""
    return map(
        list,
        zip(*sorted(
            zip(xdata, ydata),
            key=lambda x_values: x_values[0],
        )),
    )


def perform_operation(application: Graphs.Application, name: str) -> None:
    """Perform an operation."""
    this = sys.modules[__name__]
    window = application.get_window()
    if name in ("combine", ):
        return getattr(this, name)(application)
    elif name == "custom_transformation":

        def on_accept(_dialog, input_x, input_y, discard):
            try:
                _apply(application, transform, input_x, input_y, discard)
            except (RuntimeError, KeyError) as exception:
                toast = _(
                    "{name}: Unable to do transformation, \
make sure the syntax is correct",
                ).format(name=exception.__class__.__name__)
                window.add_toast_string(toast)
                logging.exception(_("Unable to do transformation"))

        dialog = Graphs.TransformDialog.new(application)
        dialog.connect("accept", on_accept)
        return
    elif name == "cut" and window.get_canvas().get_mode() != 2:
        return
    args = []
    actions_settings = application.get_settings_child("actions")
    if name in ("center", "smoothen"):
        args = [actions_settings.get_enum(name)]
    if name == "smoothen":
        params = {}
        settings = actions_settings.get_child("smoothen")
        for setting in settings:
            params[setting] = int(settings.get_int(setting))
        args += [params]
    elif name == "shift":
        figure_settings = application.get_data().get_figure_settings()
        right_range = (
            figure_settings.get_max_right() - figure_settings.get_min_right()
        )
        left_range = (
            figure_settings.get_max_left() - figure_settings.get_min_left()
        )
        args += [
            figure_settings.get_left_scale(),
            figure_settings.get_right_scale(),
            application.get_data().get_items(),
            [left_range, right_range],
        ]
    elif "translate" in name or "multiply" in name:
        try:
            args += [
                utilities.string_to_float(
                    window.get_property(name + "_entry").get_text(),
                ),
            ]
        except ValueError as error:
            window.add_toast_string(str(error))
            return
    _apply(application, getattr(this, name), *args)


def _apply(application, callback, *args):
    data = application.get_data()
    figure_settings = data.get_figure_settings()
    data_selected = False
    old_limits = figure_settings.get_limits()
    for item in data:
        if not (item.get_selected() and isinstance(item, DataItem)):
            continue
        xdata, ydata = get_data(application, item)
        if xdata is not None and len(xdata) != 0:
            data_selected = True
            new_xdata, new_ydata, sort, discard = callback(
                item, xdata, ydata, *args,
            )
            new_xdata, new_ydata = list(new_xdata), list(new_ydata)
            canvas = application.get_window().get_canvas()
            if discard and canvas.get_mode() == 2:
                logging.debug("Discard is true")
                application.get_window().add_toast_string(
                    _(
                        "Data that was outside of the highlighted area has"
                        " been discarded",
                    ),
                )
                item.props.xdata = new_xdata
                item.props.ydata = new_ydata
            else:
                logging.debug("Discard is false")
                mask = create_data_mask(
                    item.props.xdata,
                    item.props.ydata,
                    xdata,
                    ydata,
                )

                if new_xdata == []:  # If cut action was performed
                    remove_list = \
                        [index for index, masked in enumerate(mask) if masked]
                    for index in sorted(remove_list, reverse=True):
                        item.props.xdata.pop(index)
                        item.props.ydata.pop(index)
                else:
                    i = 0
                    for index, masked in enumerate(mask):
                        # Change coordinates that were within span
                        if masked:
                            item.props.xdata[index] = new_xdata[i]
                            item.props.ydata[index] = new_ydata[i]
                            i += 1
            if sort:
                logging.debug("Sorting data")
                item.xdata, item.ydata = sort_data(item.xdata, item.ydata)
        item.notify("xdata")
        item.notify("ydata")
    figure_settings.set_selection_range(0, 0)
    if not data_selected:
        application.get_window().add_toast_string(
            _("No data found within the highlighted area"),
        )
        return
    data.optimize_limits()
    data.add_history_state(old_limits)


_return = (list[float], list[float], bool, bool)


def translate_x(_item, xdata: list, ydata: list, offset: float) -> _return:
    """
    Translate all selected data on the x-axis.

    Amount to be shifted is equal to the value in the translate_x entry widget
    Will show a toast if a ValueError is raised, typically when a user entered
    an invalid number (e.g. comma instead of point separators)
    """
    return [value + offset for value in xdata], ydata, True, False


def translate_y(_item, xdata: list, ydata: list, offset: float) -> _return:
    """
    Translate all selected data on the y-axis.

    Amount to be shifted is equal to the value in the translate_y entry widget
    Will show a toast if a ValueError is raised, typically when a user entered
    an invalid number (e.g. comma instead of point separators)
    """
    return xdata, [value + offset for value in ydata], False, False


def multiply_x(_item, xdata: list, ydata: list, multiplier: float) -> _return:
    """
    Multiply all selected data on the x-axis.

    Amount to be shifted is equal to the value in the multiply_x entry widget
    Will show a toast if a ValueError is raised, typically when a user entered
    an invalid number (e.g. comma instead of point separators)
    """
    return [value * multiplier for value in xdata], ydata, True, False


def multiply_y(_item, xdata: list, ydata: list, multiplier: float) -> _return:
    """
    Multiply all selected data on the y-axis.

    Amount to be shifted is equal to the value in the multiply_y entry widget
    Will show a toast if a ValueError is raised, typically when a user entered
    an invalid number (e.g. comma instead of point separators)
    """
    return xdata, [value * multiplier for value in ydata], False, False


def normalize(_item, xdata: list, ydata: list) -> _return:
    """Normalize all selected data."""
    return xdata, [value / max(ydata) for value in ydata], False, False


def smoothen(
    _item,
    xdata: list,
    ydata: list,
    smooth_type: int,
    params: dict,
) -> None:
    """Smoothen y-data."""
    if smooth_type == 0:
        minimum = params["savgol-polynomial"] + 1
        window_percentage = params["savgol-window"] / 100
        window = max(minimum, int(len(xdata) * window_percentage))
        new_ydata = scipy.signal.savgol_filter(
            ydata,
            window,
            params["savgol-polynomial"],
        )
    elif smooth_type == 1:
        box_points = params["moving-average-box"]
        box = numpy.ones(box_points) / box_points
        new_ydata = numpy.convolve(ydata, box, mode="same")
    return xdata, new_ydata, False, False


def center(_item, xdata: list, ydata: list, center_maximum: int) -> _return:
    """
    Center all selected data.

    Depending on the key, will center either on the middle coordinate, or on
    the maximum value of the data
    """
    if center_maximum == 0:  # Center at maximum Y
        middle_index = ydata.index(max(ydata))
        middle_value = xdata[middle_index]
    elif center_maximum == 1:  # Center at middle
        middle_value = (min(xdata) + max(xdata)) / 2
    new_xdata = [coordinate - middle_value for coordinate in xdata]
    return new_xdata, ydata, True, False


def shift(
    item,
    xdata: list,
    ydata: list,
    left_scale: int,
    right_scale: int,
    items: misc.ItemList,
    ranges: tuple[float, float],
) -> _return:
    """
    Shifts data vertically with respect to each other.

    By default it scales linear data by 1.2 times the total span of the
    ydata, and log data 10 to the power of the yspan.
    """
    data_list = [
        item for item in items
        if item.get_selected() and isinstance(item, DataItem)
    ]

    y_range = ranges[1] if item.get_yposition() else ranges[0]

    shift_value_log = 0
    shift_value_linear = 0

    for index, item_ in enumerate(data_list):
        # Compare first element with itself, not "previous" item
        index = 1 if index == 0 and len(data_list) > 1 else index

        previous_item = data_list[index - 1]

        # Only use selected span when obtaining values to determine shift value
        new_xdata, new_ydata = xdata, ydata
        if (
            min(xdata) >= min(previous_item.xdata)
            and max(xdata) <= max(previous_item.xdata)
        ):
            new_xdata, new_ydata = filter_data(
                previous_item.xdata, previous_item.ydata, ">=", min(xdata))
            new_xdata, new_ydata = filter_data(
                new_xdata, new_ydata, "<=", max(xdata))

        ymin = min(x for x in new_ydata if x != 0)
        ymax = max(x for x in new_ydata if x != 0)

        scale = right_scale if item.get_yposition() else left_scale
        if scale == 1:  # Use log values for log scaling
            shift_value_log += \
                numpy.log10(abs(ymax / ymin)) + 0.1 * numpy.log10(y_range)
        else:
            shift_value_linear += (ymax - ymin) + 0.1 * y_range
        if item.get_uuid() == item_.get_uuid():
            if scale == 1:  # Log scaling
                new_ydata = [value * 10**shift_value_log for value in ydata]
            else:
                new_ydata = [value + shift_value_linear for value in ydata]
            return xdata, new_ydata, False, False
    return xdata, ydata, False, False


def cut(_item, _xdata, _ydata) -> _return:
    """Cut selected data over the span that is selected."""
    return [], [], False, False


def derivative(_item, xdata: list, ydata: list) -> _return:
    """Calculate derivative of all selected data."""
    x_values = numpy.array(xdata)
    y_values = numpy.array(ydata)
    dy_dx = numpy.gradient(y_values, x_values)
    return xdata, dy_dx.tolist(), False, True


def integral(_item, xdata: list, ydata: list) -> _return:
    """Calculate indefinite integral of all selected data."""
    x_values = numpy.array(xdata)
    y_values = numpy.array(ydata)
    indefinite_integral = scipy.integrate.cumulative_trapezoid(
        y_values,
        x_values,
        initial=0,
    ).tolist()
    return xdata, indefinite_integral, False, True


def fft(_item, xdata: list, ydata: list) -> _return:
    """Perform Fourier transformation on all selected data."""
    x_values = numpy.array(xdata)
    y_values = numpy.array(ydata)
    y_fourier = numpy.fft.fft(y_values)
    x_fourier = numpy.fft.fftfreq(len(x_values), x_values[1] - x_values[0])
    y_fourier = [value.real for value in y_fourier]
    return x_fourier, y_fourier, False, True


def inverse_fft(_item, xdata: list, ydata: list) -> _return:
    """Perform Inverse Fourier transformation on all selected data."""
    x_values = numpy.array(xdata)
    y_values = numpy.array(ydata)
    y_fourier = numpy.fft.ifft(y_values)
    x_fourier = numpy.fft.fftfreq(len(x_values), x_values[1] - x_values[0])
    y_fourier = [value.real for value in y_fourier]
    return x_fourier, y_fourier, False, True


def transform(
    _item,
    xdata: list,
    ydata: list,
    input_x: str,
    input_y: str,
    discard: bool = False,
) -> _return:
    """Perform custom transformation."""
    local_dict = {
        "x": xdata,
        "y": ydata,
        "x_min": min(xdata),
        "x_max": max(xdata),
        "y_min": min(ydata),
        "y_max": max(ydata),
    }
    # Add array of zeros to return values, such that output remains a list
    # of the correct size, even when a float is given as input.
    return (
        numexpr.evaluate(utilities.preprocess(input_x) + "+ 0*x", local_dict),
        numexpr.evaluate(utilities.preprocess(input_y) + "+ 0*y", local_dict),
        True,
        discard,
    )


def combine(application: Graphs.Application) -> None:
    """Combine the selected data into a new data set."""
    new_xdata, new_ydata = [], []
    for item in application.get_data():
        if not (item.get_selected() and isinstance(item, DataItem)):
            continue
        xdata, ydata = get_data(application, item)[:2]
        new_xdata.extend(xdata)
        new_ydata.extend(ydata)

    # Create the item itself
    new_xdata, new_ydata = sort_data(new_xdata, new_ydata)
    style = application.get_figure_style_manager().get_selected_style_params()
    application.get_data().add_items(
        [DataItem.new(style, new_xdata, new_ydata, name=_("Combined Data"))],
        application.get_figure_style_manager(),
    )



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for saving and loading projects."""
import contextlib
from gettext import gettext as _

from gi.repository import GLib, Gio, Graphs, Gtk

from graphs import file_io, migrate, misc, utilities


def read_project_file(file: Gio.File) -> dict:
    """Read a project dict from file and account for migration."""
    try:
        project_dict = file_io.parse_json(file)
    except UnicodeDecodeError:
        project_dict = migrate.migrate_project(file)
    return project_dict


def save_project_dict(file: Gio.File, project_dict: dict) -> None:
    """Save a project dict to a file."""
    file_io.write_json(file, project_dict)


_GRAPHS_PROJECT_FILE_ONLY_FILE_FILTER = utilities.create_file_filters(
    (misc.GRAPHS_PROJECT_FILE_FILTER_TEMPLATE, ),
)


def _save(application: Graphs.Application) -> None:
    data = application.get_data()
    project_file = data.props.file
    project_dict = data.get_project_dict(application.get_version())
    save_project_dict(project_file, project_dict)
    data.set_unsaved(False)
    data.emit("saved")

    application.get_window().add_toast_string_with_file(
        _("Saved Project"),
        project_file,
    )


def save_project(
    application: Graphs.Application,
    require_dialog: bool = False,
) -> None:
    """Save the current data to disk."""
    data = application.get_data()
    if data.props.file is not None and not require_dialog:
        _save(application)
        return

    def on_response(dialog, response):
        with contextlib.suppress(GLib.GError):
            data.props.file = dialog.save_finish(response)
            _save(application)

    dialog = Gtk.FileDialog()
    dialog.set_filters(_GRAPHS_PROJECT_FILE_ONLY_FILE_FILTER)
    dialog.set_initial_name(_("Project") + ".graphs")
    dialog.save(application.get_window(), None, on_response)


def open_project(application: Graphs.Application) -> None:
    """Open a project."""

    def pick_file():

        def on_pick_response(dialog, response):
            with contextlib.suppress(GLib.GError):
                load(application.get_data(), dialog.open_finish(response))

        dialog = Gtk.FileDialog()
        dialog.set_filters(_GRAPHS_PROJECT_FILE_ONLY_FILE_FILTER)
        dialog.open(application.get_window(), None, on_pick_response)

    if not application.get_data().get_unsaved():
        pick_file()
        return

    def on_save_response(_dialog, response):
        if response == "discard_close":
            pick_file()
        if response == "save_close":
            # Retrieving open dialog first means that save dialog will be
            # on top. Thus user will be presented with save dialog first.
            pick_file()
            save_project(application)

    dialog = Graphs.tools_build_dialog("save_changes")
    dialog.connect("response", on_save_response)
    dialog.present(application.get_window())


def load(data: Graphs.Data, file: Gio.File) -> None:
    """Load a project from file into data."""
    data.props.file = file
    data.load_from_project_dict(read_project_file(file))



"""Tests for operations."""
from graphs import operations
from graphs.item import DataItem

import numpy as np

import pytest

XDATA = [0, 1, 4, 5, 7, 8, 12, 1]
YDATA = [5, 2, 7, 1, 31, 5, 123, 156]


def is_sorted(lst):
    """Check if a list is sorted in ascending order."""
    return all(lst[i] <= lst[i + 1] for i in range(len(lst) - 1))


def test_sort_data():
    """Test if sort_data function sorts x and y data."""
    sorted_x, sorted_y = operations.sort_data(XDATA, YDATA)
    assert is_sorted(sorted_x)


def test_normalize():
    """Test if normalize function scales ydata to maximum value of 1."""
    xdata, ydata, _sort, _discard = operations.normalize(None, XDATA, YDATA)
    assert max(ydata) == 1


def test_translate_x():
    """Test if translate_x function correctly shifts xdata."""
    xdata, ydata, _sort, _discard = operations.translate_x(None, XDATA, YDATA,
                                                           15)
    assert all(x_new == (x_old + 15) for x_old, x_new in zip(XDATA, xdata))


def test_translate_y():
    """Test if translate_y function correctly shifts ydata."""
    xdata, ydata, _sort, _discard = operations.translate_y(None, XDATA, YDATA,
                                                           15)
    assert all(y_new == (y_old + 15) for y_old, y_new in zip(YDATA, ydata))


def test_multiply_x():
    """Test if multiply_x function correctly scales xdata."""
    xdata, ydata, _sort, _discard = operations.multiply_x(None, XDATA, YDATA,
                                                          15)
    assert all(x_new == (x_old * 15) for x_old, x_new in zip(XDATA, xdata))


def test_multiply_y():
    """Test if multiply_y function correctly scales ydata."""
    xdata, ydata, _sort, _discard = operations.multiply_y(None, XDATA, YDATA,
                                                          15)
    assert all(y_new == (y_old * 15) for y_old, y_new in zip(YDATA, ydata))


def test_center():
    """Test if center function centers ydata correctly."""
    xdata, ydata, _sort, _discard = operations.center(None, XDATA, YDATA, 0)
    y_max_index = YDATA.index(max(YDATA))
    assert xdata[y_max_index] == 0

    xdata, ydata, _sort, _discard = operations.center(None, XDATA, YDATA, 1)
    middle_value = (min(XDATA) + max(XDATA)) / 2
    assert all(x_new == (x_old - middle_value) for x_old, x_new in zip(XDATA,
                                                                       xdata))


@pytest.mark.parametrize("yscale", (0, 1, 2, 3))
def test_shift(yscale):
    """Test if shift_vertically function correctly shifts ydata."""
    ydata1 = [1.0, 1.8, 1.9, 1.1, 0.2, 0.1, 0.7, 1.7, 2.0, 1.4, 0.5]
    ydata2 = [2.0, 1.5, 0.6, 0.1, 0.3, 1.3, 2.0, 1.8, 0.9, 0.1, 0.2]
    xdata = np.linspace(0, 10, len(ydata1))

    item1 = DataItem(xdata=xdata, ydata=ydata1, uuid="a")
    item2 = DataItem(xdata=xdata, ydata=ydata2, uuid="b")

    items = [item1, item2]
    new_xdata1, new_ydata1, _sort, _discard = \
        operations.shift(item1, xdata, ydata1, left_scale=yscale,
                         right_scale=yscale, items=items, ranges=[2.2, 2.2])
    new_xdata2, new_ydata2, _sort, _discard = \
        operations.shift(item2, xdata, ydata2, left_scale=yscale,
                         right_scale=yscale, items=items, ranges=[2.2, 2.2])
    np.testing.assert_array_equal(new_xdata1, xdata)
    np.testing.assert_array_equal(new_xdata2, xdata)

    diff_y1 = [y_new - y_old for y_new, y_old in zip(new_ydata1, ydata1)]
    diff_y2 = [y_new - y_old for y_new, y_old in zip(new_ydata2, ydata2)]

    # Check that data is not removed
    assert len(new_ydata1) == len(ydata1)
    assert len(new_ydata2) == len(ydata2)
    # Check that distance increases with each shift
    assert all(
        diff_y2 > diff_y1 for new_y2, new_y1 in zip(new_ydata2, new_ydata1))
    # Check that shift is large enough for chosen coordinates
    assert all(
        new_y2 > new_y1 for new_y2, new_y1 in zip(new_ydata2, new_ydata1))


def test_derivative():
    """Test if get_derivative function correctly calculates derivative."""
    xdata = [1, 2, 3, 4, 5]
    ydata = [5, 10, 15, 10, 5]

    _, y_new, _sort, _discard = operations.derivative(None, xdata, ydata)
    assert len(y_new) == len(xdata)
    assert y_new == pytest.approx([5, 5, 0, -5, -5], rel=1e-6)


def test_integral():
    """Test if get_integral function correctly calculates integral."""
    xdata = [1, 2, 3, 4, 5]
    ydata = [5, 10, 15, 10, 5]

    x_new, y_new, _sort, _discard = operations.integral(None, xdata, ydata)

    assert len(y_new) == len(xdata)
    assert y_new == pytest.approx([0, 7.5, 20, 32.5, 40], rel=1e-6)



# SPDX-License-Identifier: GPL-3.0-or-later
"""Some UI utilities."""
import contextlib
from gettext import gettext as _, pgettext as C_

from gi.repository import GLib, Graphs, Gtk

from graphs import file_import, file_io, misc, utilities
from graphs.item_box import ItemBox


def on_items_change(
    data,
    _ignored,
    application: Graphs.Application,
) -> None:
    """Handle UI representation of items."""
    item_list = application.get_window().get_item_list()
    while item_list.get_last_child() is not None:
        item_list.remove(item_list.get_last_child())

    for index, item in enumerate(data):
        itembox = ItemBox(application, item, index)
        item_list.append(itembox)
        row = item_list.get_row_at_index(index)
        row.add_controller(itembox.get_drag_source())
        row.add_controller(itembox.get_drop_target())
        row.add_controller(itembox.get_click_gesture())
    data.add_view_history_state()


def add_data_dialog(application: Graphs.Application) -> None:
    """Show add data dialog."""

    def on_response(dialog, response):
        with contextlib.suppress(GLib.GError):
            file_import.import_from_files(
                application,
                dialog.open_multiple_finish(response),
            )

    dialog = Gtk.FileDialog()
    dialog.set_filters(
        utilities.create_file_filters((
            (
                C_("file-filter", "Supported files"),
                ["xy", "dat", "txt", "csv", "xrdml", "xry", "graphs"],
            ),
            (C_("file-filter", "ASCII files"), ["xy", "dat", "txt", "csv"]),
            (C_("file-filter", "PANalytical XRDML"), ["xrdml"]),
            (C_("file-filter", "Leybold xry"), ["xry"]),
            misc.GRAPHS_PROJECT_FILE_FILTER_TEMPLATE,
        )),
    )
    dialog.open_multiple(application.get_window(), None, on_response)


def export_data_dialog(application: Graphs.Application) -> None:
    """Show export data dialog."""
    data = application.get_data()
    window = application.get_window()
    if data.props.empty:
        window.add_toast_string(_("No data to export"))
        return
    multiple = len(data) > 1

    def on_response(dialog, response):
        with contextlib.suppress(GLib.GError):
            if multiple:
                directory = dialog.select_folder_finish(response)
                for item in data:
                    file = directory.get_child_for_display_name(
                        f"{item.get_name()}.txt",
                    )
                    file_io.save_item(file, item)
            else:
                file = dialog.save_finish(response)
                file_io.save_item(file, data[0])
            window.add_toast_string_with_file(
                _("Exported Data"),
                file,
            )

    dialog = Gtk.FileDialog()
    if multiple:
        dialog.select_folder(window, None, on_response)
    else:
        filename = f"{data[0].get_name()}.txt"
        dialog.set_initial_name(filename)
        dialog.set_filters(
            utilities.create_file_filters(
                ((C_("file-filter", "Text Files"), ["txt"]), ),
            ),
        )
        dialog.save(window, None, on_response)


def validate_entry(entry: Gtk.Editable) -> float:
    """Validate an entry for having a parsable bool value."""
    value = utilities.string_to_float(entry.get_text())
    is_valid = value is not None
    if is_valid:
        entry.remove_css_class("error")
    else:
        entry.add_css_class("error")
    return is_valid, value



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for Editing an Item."""
from gi.repository import Adw, GObject, Graphs, Gtk

from graphs.item import DataItem

_IGNORELIST = [
    "alpha",
    "color",
    "item_type",
    "uuid",
    "selected",
    "xdata",
    "xlabel",
    "ydata",
    "ylabel",
]


@Gtk.Template(resource_path="/se/sjoerd/Graphs/ui/edit-item.ui")
class EditItemDialog(Adw.PreferencesDialog):
    """Class for editing an Item."""

    __gtype_name__ = "GraphsEditItemDialog"
    item_selector = Gtk.Template.Child()
    name = Gtk.Template.Child()
    xposition = Gtk.Template.Child()
    yposition = Gtk.Template.Child()

    item_group = Gtk.Template.Child()
    linestyle = Gtk.Template.Child()
    linewidth = Gtk.Template.Child()
    markerstyle = Gtk.Template.Child()
    markersize = Gtk.Template.Child()

    data = GObject.Property(type=Graphs.Data)
    item = GObject.Property(type=Graphs.Item)
    model = GObject.Property(type=Gtk.StringList)
    bindings = GObject.Property(type=object)

    def __init__(self, application, item):
        data = application.get_data()
        super().__init__(
            data=data,
            item=item,
            bindings=[],
            model=Gtk.StringList.new(data.get_names()),
        )
        self.item_selector.set_model(self.props.model)
        self.item_selector.set_selected(data.get_items().index(item))
        self.present(application.get_window())

    @Gtk.Template.Callback()
    def on_select(self, _action, _target) -> None:
        """Handle Item selection."""
        item = self.props.data[self.item_selector.get_selected()]
        if item != self.item:
            self.props.model.splice(
                self.props.data.get_items().index(self.item),
                1,
                [self.props.item.get_name()],
            )
            self.props.item = item

    @Gtk.Template.Callback()
    def on_item_change(self, _a, _b) -> None:
        """Handle Item change."""
        item = self.props.item
        self.set_title(item.get_name())
        for binding in self.props.bindings:
            binding.unbind()
        bindings = []
        for key in dir(item.props):
            if key in _IGNORELIST:
                continue
            widget = getattr(self, key)
            if isinstance(widget, Adw.EntryRow):
                bindings.append(
                    item.bind_property(key, widget, "text", 1 | 2),
                )
            elif isinstance(widget, Adw.ComboRow):
                bindings.append(
                    item.bind_property(key, widget, "selected", 1 | 2),
                )
            elif isinstance(widget, Gtk.Scale):
                bindings.append(
                    item.bind_property(
                        key,
                        widget.get_adjustment(),
                        "value",
                        1 | 2,
                    ),
                )
        self.props.bindings = bindings
        self.item_group.set_visible(isinstance(item, DataItem))

    @Gtk.Template.Callback()
    def on_close(self, _a) -> None:
        """Handle dialog closing."""
        self.props.data.add_history_state()

    @Gtk.Template.Callback()
    def on_linestyle(self, comborow, _b) -> None:
        """Handle linestyle selection."""
        self.linewidth.set_sensitive(comborow.get_selected() != 0)

    @Gtk.Template.Callback()
    def on_markers(self, comborow, _b) -> None:
        """Handle marker selection."""
        self.markersize.set_sensitive(comborow.get_selected() != 0)



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Custom canvas implementation.

Acts as an interface between matplotlib and GObject and contains a custom
implementation of a `SpanSelector` as well as a dummy toolbar used for
interactive navigation in conjunction with graphs-specific structures.

    Classes:
        Canvas
"""
import math

from gi.repository import Adw, GObject, Gdk, Graphs, Gtk

from graphs import artist, misc, scales, utilities

from matplotlib import backend_tools as tools, pyplot
from matplotlib.backend_bases import (
    FigureCanvasBase,
    MouseEvent,
    NavigationToolbar2,
)
from matplotlib.backends.backend_gtk4cairo import FigureCanvas
from matplotlib.widgets import SpanSelector

_SCROLL_SCALE = 1.06


class Canvas(Graphs.Canvas, FigureCanvas):
    """
    Custom Canvas.

    Implements properties analouge to `FigureSettings`. Automatically connects
    to `FigureSettings` and `Data.items` during init.

    Properties:
        mode: int

        items: list (write-only)

        title: str
        bottom_label: str
        left_label: str
        top_label: str
        right_label: str

        bottom_scale: int (0: linear, 1: logarithmic)
        left_scale: int (0: linear, 1: logarithmic)
        top_scale: int (0: linear, 1: logarithmic)
        right_scale: int (0: linear, 1: logarithmic)

        legend: bool
        legend_position: int
        use_custom_style: bool
        custom_style: str

        min_bottom: float
        max_bottom: float
        min_left: float
        max_lef: float
        min_top: float
        max_top: float
        min_right: float
        max_right: float

        min_selected: float (fraction)
        max_selected: float (fraction)

    Signals:
        edit_request
        view_changed

    Functions:
        update_legend
    """

    __gtype_name__ = "GraphsPythonCanvas"

    items = GObject.Property(type=object)

    def __init__(
        self,
        style_params: dict,
        interactive: bool = True,
    ):
        """
        Create the canvas.

        Create figure, axes and define rubberband_colors based on the current
        style context. Bind `items` to `data.items` and all figure settings
        attributes to their respective values.
        """
        self._style_params = style_params
        pyplot.rcParams.update(self._style_params)  # apply style_params
        Graphs.Canvas.__init__(
            self,
            can_focus=True,
            focusable=True,
            hexpand=True,
            vexpand=True,
            items=[],
        )
        self._idle_draw_id = 0
        self.set_draw_func(self._draw_func)
        self.connect("resize", self.resize_event)
        self.connect("notify::scale-factor", self._update_device_pixel_ratio)
        FigureCanvasBase.__init__(self)
        self.figure.set_tight_layout(True)
        self._axis = self.figure.add_subplot(111)
        self._top_left_axis = self._axis.twiny()
        self._right_axis = self._axis.twinx()
        self._top_right_axis = self._top_left_axis.twinx()
        self.axes = [
            self._axis,
            self._top_left_axis,
            self._right_axis,
            self._top_right_axis,
        ]
        self._legend_axis = self._axis
        self._legend = True
        self._legend_position = misc.LEGEND_POSITIONS[0]
        self._handles = []
        self._rubberband_rect = None

        # Handle stuff only used if the canvas is interactive
        if interactive:
            self._setup_interactive()

        self.connect("notify::hide-unselected", self._redraw)
        self.connect("notify::items", self._redraw)

    def _setup_interactive(self):
        self._ctrl_held, self._shift_held = False, False
        self._xfrac, self._yfrac = None, None
        self.mpl_connect("pick_event", self._on_pick)
        self.mpl_connect("motion_notify_event", self._set_mouse_fraction)

        # Reference is created by the toolbar itself
        _DummyToolbar(self)

        click = Gtk.GestureClick()
        click.set_button(0)  # All buttons.
        click.connect("pressed", self.button_press_event)
        click.connect("update", self.handle_touch_update)
        click.connect("released", self.button_release_event)
        self.add_controller(click)

        key = Gtk.EventControllerKey()
        key.connect("key-pressed", self.key_press_event)
        key.connect("key-released", self.key_release_event)
        self.add_controller(key)

        motion = Gtk.EventControllerMotion()
        motion.connect("motion", self.motion_notify_event)
        motion.connect("enter", self.enter_notify_event)
        motion.connect("leave", self.leave_notify_event)
        self.add_controller(motion)

        scroll = Gtk.EventControllerScroll.new(
            Gtk.EventControllerScrollFlags.BOTH_AXES,
        )
        scroll.connect("scroll", self.scroll_event)
        scroll.connect("scroll-end", self.toolbar.push_current)
        self.add_controller(scroll)

        zoom = Gtk.GestureZoom.new()
        zoom.connect("scale-changed", self.zoom_event)
        zoom.connect("end", self.end_zoom_event)
        self.add_controller(zoom)

        def rgba_to_tuple(rgba):
            return (rgba.red, rgba.green, rgba.blue, rgba.alpha)

        style_manager = Adw.StyleManager()
        rgba = style_manager.get_accent_color_rgba()
        self.rubberband_edge_color = rgba_to_tuple(rgba)
        rgba.alpha = 0.2
        self.rubberband_fill_color = rgba_to_tuple(rgba)
        self.highlight = _Highlight(self)

        for item in ("min", "max"):
            self.connect(
                f"notify::{item}-selected",
                lambda _a,
                _b: self.highlight.load(self),
            )

    def handle_touch_update(self, controller: Gtk.GestureClick, _data) -> None:
        """
        Handle an update event for GtkGestureClick motion.

        This is needed for touch screen devices to handle gestures properly.
        """
        if not controller.get_point()[0]:  # If touch event
            coords = controller.get_bounding_box_center()
            x, y = coords.x, coords.y
            MouseEvent(
                "motion_notify_event",
                self,
                *self._mpl_coords((x, y)),
            )._process()

    def key_press_event(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> None:
        """Handle key press event."""
        if keyval == 65507 or keyval == 65508:  # Control_L or Control_R
            self._ctrl_held = True
        elif keyval == 65505 or keyval == 65506:  # Left or right Shift
            self._shift_held = True
        else:  # Prevent keys from being true with key combos
            self._ctrl_held = False
        super().key_press_event(controller, keyval, keycode, state)

    def key_release_event(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> None:
        """Handle key release event."""
        self._ctrl_held = False
        self._shift_held = False
        super().key_release_event(controller, keyval, keycode, state)

    def scroll_event(
        self,
        controller: Gtk.EventControllerScroll,
        dx: float,
        dy: float,
    ) -> None:
        """Handle scroll event."""
        if self._ctrl_held:
            self.zoom(1 / _SCROLL_SCALE if dy > 0 else _SCROLL_SCALE)
        else:
            if self._shift_held:
                dx, dy = dy, dx
            if controller.get_unit() == Gdk.ScrollUnit.WHEEL:
                dx *= 10
                dy *= 10
            for ax in self.axes:
                xmin, xmax, ymin, ymax = \
                    self._calculate_pan_values(ax, dx, dy)
                ax.set_xlim(xmin, xmax)
                ax.set_ylim(ymin, ymax)
            self.queue_draw()
        super().scroll_event(controller, dx, dy)

    def zoom_event(
        self,
        controller: Gtk.GestureZoom,
        scale: float,
    ) -> None:
        """Handle zoom event."""
        coords = controller.get_bounding_box_center()
        x, y = coords.x, coords.y
        event = MouseEvent(
            "motion_notify_event",
            self,
            *self._mpl_coords((x, y)),
        )
        self._set_mouse_fraction(event)

        scale = 1 + 0.01 * (scale - 1)
        if scale > 5 or scale < 0.2:
            # Don't scale if ridiculous values are registered
            return
        self.zoom(scale)

    def end_zoom_event(self, controller: Gtk.GestureZoom, _sequence) -> None:
        """
        End the zoom event.

        Pushes the canvas to the stack, and emits a `release` signal cancel out
        registered touches from touchscreen devices.
        """
        coords = controller.get_bounding_box_center()
        x, y = coords.x, coords.y
        MouseEvent(
            "button_release_event",
            self,
            *self._mpl_coords((x, y)),
            1,
        )._process()
        self.toolbar.push_current()

    def enter_notify_event(
        self,
        controller: Gtk.EventControllerMotion,
        x: float,
        y: float,
    ) -> None:
        """Process pointer entry."""
        self.grab_focus()
        super().enter_notify_event(controller, x, y)

    def _set_mouse_fraction(self, event) -> None:
        """Set the mouse coordinate in terms of fraction of the canvas."""
        if event.inaxes is not None:
            xlim = self._top_right_axis.get_xlim()
            ylim = self._top_right_axis.get_ylim()
            self._xfrac = utilities.get_fraction_at_value(
                event.xdata,
                xlim[0],
                xlim[1],
                self.top_scale,
            )
            self._yfrac = utilities.get_fraction_at_value(
                event.ydata,
                ylim[0],
                ylim[1],
                self.right_scale,
            )
        else:
            self._xfrac, self._yfrac = None, None

    def zoom(self, scaling: float = 1.15, respect_mouse: bool = True) -> None:
        """
        Zoom with given scaling.

        Update all axes' limits in respect to the current mouse position.
        """
        if not respect_mouse:
            self._xfrac, self._yfrac = 0.5, 0.5
        if self._xfrac is None or self._yfrac is None:
            return
        for ax in self.axes:
            ax.set_xlim(
                self._calculate_zoomed_values(
                    self._xfrac,
                    scales.to_int(ax.get_xscale()),
                    ax.get_xlim(),
                    scaling,
                ),
            )
            ax.set_ylim(
                self._calculate_zoomed_values(
                    self._yfrac,
                    scales.to_int(ax.get_yscale()),
                    ax.get_ylim(),
                    scaling,
                ),
            )
        self.queue_draw()

    @staticmethod
    def _calculate_pan_values(
        ax: pyplot.axis,
        x_panspeed: float,
        y_panspeed: float,
    ) -> None:
        """
        Calculate values required for panning.

        Calculates the coordinates of the canvas after a panning gesture has
        been emitted.
        """
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        x_scale = scales.to_int(ax.get_xscale())
        y_scale = scales.to_int(ax.get_yscale())
        pan_scale = 0.002
        xvalue1 = utilities.get_value_at_fraction(
            x_panspeed * pan_scale,
            xmin,
            xmax,
            x_scale,
        )
        xvalue2 = utilities.get_value_at_fraction(
            1 + x_panspeed * pan_scale,
            xmin,
            xmax,
            x_scale,
        )
        yvalue1 = utilities.get_value_at_fraction(
            -y_panspeed * pan_scale,
            ymin,
            ymax,
            y_scale,
        )
        yvalue2 = utilities.get_value_at_fraction(
            1 - y_panspeed * pan_scale,
            ymin,
            ymax,
            y_scale,
        )
        if x_scale == 4:
            xvalue1, xvalue2 = xvalue2, xvalue1
        if y_scale == 4:
            yvalue1, yvalue2 = yvalue2, yvalue1
        return xvalue1, xvalue2, yvalue1, yvalue2

    @staticmethod
    def _calculate_zoomed_values(
        fraction: float,
        scale: float,
        limit: float,
        zoom_factor: float,
    ) -> tuple[float, float]:
        """
        Calculate zoomed values.

        Calculates the coordinates of the canvas after a zoom gesture
        has  been ezoomed.
        """
        min_, max_ = limit[0], limit[1]
        value1 = utilities.get_value_at_fraction(
            fraction - fraction / zoom_factor,
            min_,
            max_,
            scale,
        )
        value2 = utilities.get_value_at_fraction(
            fraction + (1 - fraction) / zoom_factor,
            min_,
            max_,
            scale,
        )
        if scale == 4:
            value1, value2 = value2, value1
        return value1, value2

    def _redraw(self, *_args) -> None:
        # bottom, top, left, right
        used_axes = [False, False, False, False]
        visible_axes = [False, False, False, False]
        if self.props.hide_unselected:
            drawable_items = []
            for item in self.props.items:
                if hasattr(item, "handler"):
                    item.disconnect(item.handler)
                item.handler = item.connect("notify::selected", self._redraw)
                if item.get_selected():
                    drawable_items.append(item)
        else:
            drawable_items = self.props.items
        for item in drawable_items:
            xposition = item.get_xposition()
            yposition = item.get_yposition()
            visible_axes[xposition] = True
            visible_axes[2 + yposition] = True
            used_axes[xposition + 2 * yposition] = True
        axes_directions = (
            ("bottom", "left"),  # axis
            ("top", "left"),  # top_left_axis
            ("bottom", "right"),  # right_axis
            ("top", "right"),  # top_right_axis
        )

        if not any(visible_axes):
            visible_axes = (True, False, True, False)  # Left and bottom
            used_axes = (True, False, False, False)  # self.axis visible
            self._legend_axis = self._axis

        params = self._style_params
        draw_frame = params["axes.spines.bottom"]
        ticks = "both" if params["xtick.minor.visible"] else "major"
        for directions, axis, used \
                in zip(axes_directions, self.axes, used_axes):
            axis.get_xaxis().set_visible(False)
            axis.get_yaxis().set_visible(False)
            # Set tick where requested, as long as that axis is not occupied
            # and visible
            if (
                params[f"xtick.{directions[0]}"]
                or params[f"ytick.{directions[1]}"]
            ):
                axis.tick_params(
                    which=ticks,
                    **{
                        direction: (
                            draw_frame and not visible_axes[i]
                            or direction in directions
                        )
                        and params[f"{'x' if i < 2 else 'y'}tick.{direction}"]
                        for i,
                        direction in enumerate(misc.DIRECTIONS)
                    },
                )
            for handle in axis.lines + axis.texts:
                handle.remove()
            axis_legend = axis.get_legend()
            if axis_legend is not None:
                axis_legend.remove()
            for direction in misc.DIRECTIONS:
                axis.spines[direction].set_visible(
                    direction in directions and used or draw_frame,
                )
            if used:
                self._legend_axis = axis

        self._axis.get_xaxis().set_visible(visible_axes[0])
        self._top_left_axis.get_xaxis().set_visible(visible_axes[1])
        self._axis.get_yaxis().set_visible(visible_axes[2])
        self._right_axis.get_yaxis().set_visible(visible_axes[3])

        self._handles = [
            artist.new_for_item(self, item)
            for item in reversed(drawable_items)
        ]
        self.update_legend()

    def _on_pick(self, event) -> None:
        """Emit edit-request signal for picked label/title."""
        self.emit("edit_request", event.artist.id)

    # Overwritten function - do not change name
    def _post_draw(self, _widget, context) -> None:
        """Allow custom rendering extensions."""
        if self._rubberband_rect is not None:
            self._draw_rubberband(context)

    def _draw_rubberband(self, ctx) -> None:
        """
        Implement custom rubberband.

        Draw a rubberband matching libadwaitas style, where `_rubberband_rect`
        is set.
        """
        radius = 6
        x0, y0, width, height = (
            dim / self.device_pixel_ratio for dim in self._rubberband_rect
        )
        x1 = x0 + width
        y1 = y0 + height

        if x1 < x0:
            x0, x1 = x1, x0
        if y1 < y0:
            y0, y1 = y1, y0

        degrees = math.pi / 180

        ctx.new_sub_path()
        ctx.arc(x1 - radius, y0 + radius, radius, -90 * degrees, 0 * degrees)
        ctx.arc(x1 - radius, y1 - radius, radius, 0 * degrees, 90 * degrees)
        ctx.arc(x0 + radius, y1 - radius, radius, 90 * degrees, 180 * degrees)
        ctx.arc(x0 + radius, y0 + radius, radius, 180 * degrees, 270 * degrees)
        ctx.close_path()

        ctx.set_line_width(1)
        ctx.set_source_rgba(*self.rubberband_fill_color)
        ctx.fill_preserve()
        ctx.set_source_rgba(*self.rubberband_edge_color)
        ctx.stroke()

    def update_legend(self) -> None:
        """Update the legend or hide if not used."""
        if self._legend and self._handles:
            handles = [
                handle.get_artist() for handle in self._handles
                if handle.legend
            ]
            if handles:
                self._legend_axis.legend(
                    handles=handles,
                    loc=self._legend_position,
                    frameon=True,
                    reverse=True,
                )
                self.queue_draw()
                return
        legend = self._legend_axis.get_legend()
        if legend is not None:
            legend.remove()
        self.queue_draw()

    @GObject.Property(type=bool, default=True)
    def legend(self) -> bool:
        """Whether or not, the legend is visible."""
        return self._legend

    @legend.setter
    def legend(self, legend: bool) -> None:
        self._legend = legend
        self.update_legend()

    @GObject.Property(type=int, default=0)
    def legend_position(self) -> int:
        """Legend Position (see `misc.LEGEND_POSITIONS`)."""
        return misc.LEGEND_POSITIONS.index(self._legend_position)

    @legend_position.setter
    def legend_position(self, legend_position: int) -> None:
        self._legend_position = misc.LEGEND_POSITIONS[legend_position]
        self.update_legend()

    @GObject.Property(type=str)
    def title(self) -> str:
        """Figure title."""
        return self._axis.get_title()

    @title.setter
    def title(self, title: str) -> None:
        self._axis.set_title(title, picker=True).id = "title"
        self.queue_draw()

    @GObject.Property(type=str)
    def bottom_label(self) -> str:
        """Label of the bottom axis."""
        return self._axis.get_xlabel()

    @bottom_label.setter
    def bottom_label(self, label: str) -> None:
        self._axis.set_xlabel(label, picker=True).id = "bottom_label"
        self.queue_draw()

    @GObject.Property(type=str)
    def left_label(self) -> str:
        """Label of the left axis."""
        return self._axis.get_ylabel()

    @left_label.setter
    def left_label(self, label: str) -> None:
        self._axis.set_ylabel(label, picker=True).id = "left_label"
        self.queue_draw()

    @GObject.Property(type=str)
    def top_label(self) -> str:
        """Label of the top axis."""
        return self._top_left_axis.get_xlabel()

    @top_label.setter
    def top_label(self, label: str) -> None:
        self._top_left_axis.set_xlabel(label, picker=True).id = "top_label"
        self.queue_draw()

    @GObject.Property(type=str)
    def right_label(self) -> str:
        """Label of the right axis."""
        return self._right_axis.get_ylabel()

    @right_label.setter
    def right_label(self, label: str) -> None:
        self._right_axis.set_ylabel(label, picker=True).id = "right_label"
        self.queue_draw()

    @GObject.Property(type=int)
    def bottom_scale(self) -> int:
        """Scale of the bottom axis."""
        return scales.to_int(self._axis.get_xscale())

    @bottom_scale.setter
    def bottom_scale(self, scale: int) -> None:
        scale = scales.to_string(scale)
        for axis in (self._axis, self._right_axis):
            axis.set_xscale(scale)
            axis.set_xlim(None, None)
        self.queue_draw()

    @GObject.Property(type=int)
    def left_scale(self) -> int:
        """Scale of the left axis."""
        return scales.to_int(self._axis.get_yscale())

    @left_scale.setter
    def left_scale(self, scale: int) -> None:
        scale = scales.to_string(scale)
        for axis in (self._axis, self._top_left_axis):
            axis.set_yscale(scale)
            axis.set_ylim(None, None)
        self.queue_draw()

    @GObject.Property(type=int)
    def top_scale(self) -> int:
        """Scale of the top axis."""
        return scales.to_int(self._top_left_axis.get_xscale())

    @top_scale.setter
    def top_scale(self, scale: int) -> None:
        scale = scales.to_string(scale)
        for axis in (self._top_right_axis, self._top_left_axis):
            axis.set_xscale(scale)
            axis.set_xlim(None, None)
        self.queue_draw()

    @GObject.Property(type=int)
    def right_scale(self) -> int:
        """Scale of the right axis."""
        return scales.to_int(self._right_axis.get_yscale())

    @right_scale.setter
    def right_scale(self, scale: int) -> None:
        scale = scales.to_string(scale)
        for axis in (self._top_right_axis, self._right_axis):
            axis.set_yscale(scale)
            axis.set_ylim(None, None)
        self.queue_draw()

    @GObject.Property(type=float)
    def min_bottom(self) -> float:
        """Lower limit for the bottom axis."""
        return self._axis.get_xlim()[0]

    @min_bottom.setter
    def min_bottom(self, value: float) -> None:
        for axis in (self._axis, self._right_axis):
            axis.set_xlim(value, None)
        self.queue_draw()

    @GObject.Property(type=float)
    def max_bottom(self) -> float:
        """Upper limit for the bottom axis."""
        return self._axis.get_xlim()[1]

    @max_bottom.setter
    def max_bottom(self, value: float) -> None:
        for axis in (self._axis, self._right_axis):
            axis.set_xlim(None, value)
        self.queue_draw()

    @GObject.Property(type=float)
    def min_left(self) -> float:
        """Lower limit for the left axis."""
        return self._axis.get_ylim()[0]

    @min_left.setter
    def min_left(self, value: float) -> None:
        for axis in (self._axis, self._top_left_axis):
            axis.set_ylim(value, None)
        self.queue_draw()

    @GObject.Property(type=float)
    def max_left(self) -> float:
        """Upper limit for the left axis."""
        return self._axis.get_ylim()[1]

    @max_left.setter
    def max_left(self, value: float) -> None:
        for axis in (self._axis, self._top_left_axis):
            axis.set_ylim(None, value)
        self.queue_draw()

    @GObject.Property(type=float)
    def min_top(self) -> float:
        """Lower limit for the top axis."""
        return self._top_left_axis.get_xlim()[0]

    @min_top.setter
    def min_top(self, value: float) -> None:
        for axis in (self._top_left_axis, self._top_right_axis):
            axis.set_xlim(value, None)
        self.highlight.load(self)
        self.queue_draw()

    @GObject.Property(type=float)
    def max_top(self) -> float:
        """Upper limit for the top axis."""
        return self._top_left_axis.get_xlim()[1]

    @max_top.setter
    def max_top(self, value: float) -> None:
        for axis in (self._top_left_axis, self._top_right_axis):
            axis.set_xlim(None, value)
        self.highlight.load(self)
        self.queue_draw()

    @GObject.Property(type=float)
    def min_right(self) -> float:
        """Lower limit for the right axis."""
        return self._right_axis.get_ylim()[0]

    @min_right.setter
    def min_right(self, value: float) -> None:
        for axis in (self._right_axis, self._top_right_axis):
            axis.set_ylim(value, None)
        self.queue_draw()

    @GObject.Property(type=float)
    def max_right(self) -> float:
        """Upper limit for the right axis."""
        return self._right_axis.get_ylim()[1]

    @max_right.setter
    def max_right(self, value: float) -> None:
        for axis in (self._right_axis, self._top_right_axis):
            axis.set_ylim(None, value)
        self.queue_draw()

    @GObject.Property(type=bool, default=False)
    def highlight_enabled(self) -> bool:
        """Whether or not the highlight is enabled."""
        return hasattr(self, "highlight") and self.highlight.get_active()

    @highlight_enabled.setter
    def highlight_enabled(self, enabled: bool) -> None:
        if hasattr(self, "highlight"):
            self.highlight.set_active(enabled)
            self.highlight.set_visible(enabled)
            self.queue_draw()


class _DummyToolbar(NavigationToolbar2):
    """Custom Toolbar implementation."""

    # Overwritten function - do not change name
    def _zoom_pan_handler(self, event) -> None:
        mode = self.canvas.props.mode
        if event.button == 2:
            event.button = 1
            mode = 0
        elif event.button != 1:
            return
        if mode == 0:
            if event.name == "button_press_event":
                self.press_pan(event)
            elif event.name == "button_release_event":
                self.release_pan(event)
        elif mode == 1:
            if event.name == "button_press_event":
                self.press_zoom(event)
            elif event.name == "button_release_event":
                self.release_zoom(event)

    # Overwritten function - do not change name
    def _update_cursor(self, event) -> None:
        mode = self.canvas.props.mode
        if event.inaxes and event.inaxes.get_navigate():
            if mode == 1 and self._last_cursor != tools.Cursors.SELECT_REGION:
                self.canvas.set_cursor(tools.Cursors.SELECT_REGION)
                self._last_cursor = tools.Cursors.SELECT_REGION
            elif mode == 0 and self._last_cursor != tools.Cursors.MOVE:
                self.canvas.set_cursor(tools.Cursors.MOVE)
                self._last_cursor = tools.Cursors.MOVE
        elif self._last_cursor != tools.Cursors.POINTER:
            self.canvas.set_cursor(tools.Cursors.POINTER)
            self._last_cursor = tools.Cursors.POINTER

    # Overwritten function - do not change name
    def drag_pan(self, event):
        """Handle dragging in pan/zoom mode."""
        for ax in self._pan_info.axes:
            # Using the recorded button at the press is safer than the current
            # button, as multiple buttons can get pressed during motion.
            # Use custom drag_pan that maxes sure limits are set in right order
            # even on inverted scale
            self.ax_drag_pan(
                ax,
                self._pan_info.button,
                event.key,
                event.x,
                event.y,
            )
        self.canvas.draw_idle()

    @staticmethod
    def ax_drag_pan(self, button, key: str, x: float, y: float) -> None:
        """
        Handle mouse events during a pan operation.

        Notes
        -----
        This is intended to be overridden by new projection types.
        """
        points = self._get_pan_points(button, key, x, y)
        if points is not None:
            # Max and min needs to be defined at correct position for this to
            # work with inverted scaling
            ylim = points[:, 1]
            xlim = points[:, 0]
            self.set_xlim(min(xlim), max(xlim))
            self.set_ylim(min(ylim), max(ylim))

    # Overwritten function - do not change name
    def draw_rubberband(self, _event, x0, y0, x1, y1) -> None:
        self.canvas._rubberband_rect = [
            int(val) for val in
            (x0, self.canvas.figure.bbox.height - y0, x1 - x0, y0 - y1)
        ]
        self.canvas.queue_draw()

    # Overwritten function - do not change name
    def remove_rubberband(self) -> None:
        self.canvas._rubberband_rect = None
        self.canvas.queue_draw()

    # Overwritten function - do not change name
    def push_current(self, *_args) -> None:
        """Use custom functionality for the view clipboard."""
        self.canvas.highlight.load(self.canvas)
        for direction in ("bottom", "left", "top", "right"):
            self.canvas.notify(f"min-{direction}")
            self.canvas.notify(f"max-{direction}")
        self.canvas.emit("view_changed")

    # Overwritten function - do not change name
    def save_figure(self) -> None:
        pass


class _Highlight(SpanSelector):

    def __init__(self, canvas: Canvas):
        super().__init__(
            canvas.axes[3],
            lambda _x,
            _y: self.apply(canvas),
            "horizontal",
            useblit=True,
            props={
                "facecolor": canvas.rubberband_fill_color,
                "edgecolor": canvas.rubberband_edge_color,
                "linewidth": 1,
            },
            handle_props={"linewidth": 0},
            interactive=True,
            drag_from_anywhere=True,
        )
        self.load(canvas)

    def load(self, canvas: Canvas) -> None:
        xmin, xmax = canvas.axes[1].get_xlim()
        scale = canvas.props.top_scale
        self.extents = (
            utilities.get_value_at_fraction(
                canvas.get_min_selected(),
                xmin,
                xmax,
                scale,
            ),
            utilities.get_value_at_fraction(
                canvas.get_max_selected(),
                xmin,
                xmax,
                scale,
            ),
        )

    def apply(self, canvas: Canvas) -> None:
        xmin, xmax = canvas.axes[1].get_xlim()
        extents = self.extents
        extents = max(xmin, extents[0]), min(xmax, extents[1])
        self.extents = extents

        for prefix, value in zip(["min_", "max_"], extents):
            canvas.set_property(
                prefix + "selected",
                utilities.get_fraction_at_value(
                    value,
                    xmin,
                    xmax,
                    canvas.props.top_scale,
                ),
            )



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for style utilities."""
import contextlib
import io
import os
from gettext import gettext as _
from pathlib import Path

from cycler import cycler

from gi.repository import (
    Adw,
    GLib,
    GObject,
    Gdk,
    Gio,
    Graphs,
    Gtk,
    Pango,
)

import graphs
from graphs import item, style_io

from matplotlib import RcParams, rcParams, rcParamsDefault


def _generate_filename(name: str) -> str:
    name = name.replace("(", "").replace(")", "")
    return f"{name.lower().replace(' ', '-')}.mplstyle"


def _is_style_bright(params: RcParams):
    return Graphs.tools_get_luminance_from_hex(params["axes.facecolor"]) < 0.4


def _generate_preview(style: RcParams) -> Gdk.Texture:
    buffer = io.BytesIO()
    style_io.create_preview(buffer, style)
    return Gdk.Texture.new_from_bytes(GLib.Bytes.new(buffer.getvalue()))


class StyleManager(Graphs.StyleManager):
    """
    Main Style Manager.

    Keeps track of all files in the style dir and represents them in
    the `selection_model` property.
    """

    __gtype_name__ = "GraphsPythonStyleManager"

    def __init__(self, application: Graphs.Application):
        # Check for Ubuntu
        gtk_theme = Gtk.Settings.get_default().get_property("gtk-theme-name")
        self._system_style_name = "Yaru" \
            if "SNAP" in os.environ \
            and gtk_theme.lower().startswith("yaru") \
            else "Adwaita"
        super().__init__(application=application)
        self._selected_style_params = None
        self._stylenames = self.load_system_styles(
            self._system_style_name.lower(),
        )
        self._update_system_style()
        enumerator = self.props.style_dir.enumerate_children(
            "default::*",
            0,
            None,
        )
        for file in map(enumerator.get_child, enumerator):
            if file.query_file_type(0, None) != 1:
                continue
            if Path(Graphs.tools_get_filename(file)).suffix != ".mplstyle":
                continue
            self._add_user_style(file)
        enumerator.close(None)
        self._style_monitor = self.props.style_dir.monitor_directory(0, None)
        self._style_monitor.connect("changed", self._on_file_change)

        application.get_style_manager().connect(
            "notify",
            self._on_system_style_change,
        )
        notifiers = ("custom_style", "use_custom_style")
        for prop in notifiers:
            self.connect(
                "notify::" + prop.replace("_", "-"),
                getattr(self, "_on_" + prop),
            )

        self.setup_bindings(application.get_data().get_figure_settings())
        self._on_style_change()

    def _on_system_style_change(self, _a, _b):
        if not self.props.use_custom_style:
            self._on_style_change()

    def _add_user_style(
        self,
        file: Gio.File,
        style_params: RcParams = None,
        name: str = None,
    ) -> None:
        if style_params is None:
            tmp_style_params, name = style_io.parse(file)
            style_params = self._complete_style(tmp_style_params)
        if name in self._stylenames:
            new_name = Graphs.tools_get_duplicate_string(
                name,
                self._stylenames,
            )
            file.delete(None)
            file = self.props.style_dir.get_child_for_display_name(
                _generate_filename(new_name),
            )
            style_io.write(style_params, new_name, file)
        self._stylenames.append(name)
        self.props.selection_model.get_model().insert_sorted(
            Graphs.Style(
                name=name,
                file=file,
                mutable=True,
                preview=_generate_preview(style_params),
                light=_is_style_bright(style_params),
            ),
            Graphs.style_cmp,
        )
        if not self.props.use_custom_style:
            return

        old_style = self.props.custom_style
        if old_style not in self._stylenames:
            self.props.custom_style = name

        for index, style in enumerate(self.props.selection_model):
            if index > 0 and style.get_name() == self.props.custom_style:
                self.props.selection_model.set_selected(index)
                break

    def get_selected_style_params(self) -> RcParams:
        """Get the selected style properties."""
        return self._selected_style_params

    def get_system_style_params(self) -> RcParams:
        """Get the system style properties."""
        return self._system_style_params

    def _on_file_change(
        self,
        _monitor,
        file: Gio.File,
        _other_file,
        event_type: int,
    ) -> None:
        if Path(file.peek_path()).stem.startswith("."):
            return
        possible_visual_impact = False
        stylename = None
        style_model = self.props.selection_model.get_model()
        if event_type == 2:
            for index, style in enumerate(style_model):
                file2 = style.get_file()
                if file2 is not None and file.equal(file2):
                    stylename = style.get_name()
                    self._stylenames.remove(stylename)
                    style_model.remove(index)
                    break
            if stylename is None:
                return
            possible_visual_impact = True
        else:
            tmp_style_params, stylename = style_io.parse(file)
            style_params = self._complete_style(tmp_style_params)
        if event_type == 1:
            for obj in style_model:
                if obj.get_name() == stylename:
                    obj.set_preview(_generate_preview(style_params))
                    obj.set_light(_is_style_bright(style_params))
                    break
            possible_visual_impact = False
        elif event_type == 3:
            self._add_user_style(file, style_params, stylename)
        if possible_visual_impact \
                and self.props.use_custom_style \
                and self.props.custom_style == stylename \
                and event_type != 2:
            self._on_style_change()

    def _on_style_select(self, style_model, _pos, _n_items):
        """Set the style upon selection."""
        selected_item = style_model.get_selected_item()
        # Don't trigger unneccesary reloads
        if selected_item.get_file() is None:  # System style
            if self.props.use_custom_style:
                self.props.use_custom_style = False
        else:
            stylename = selected_item.get_name()
            if stylename != self.props.custom_style:
                self.props.custom_style = stylename
            if not self.props.use_custom_style:
                self.props.use_custom_style = True

    @staticmethod
    def _on_use_custom_style(self, _a) -> None:
        """Handle `use_custom_style` property change."""
        if self.props.use_custom_style:
            self._on_custom_style(self, None)
        else:
            self.props.selection_model.set_selected(0)
        self._on_style_change(True)

    @staticmethod
    def _on_custom_style(self, _a) -> None:
        """Handle `custom_style` property change."""
        if self.props.use_custom_style:
            for index, style in enumerate(self.props.selection_model):
                if index > 0 and style.get_name() == self.props.custom_style:
                    self.props.selection_model.set_selected(index)
                    break
            self._on_style_change(True)

    def _on_style_change(self, override: bool = False) -> None:
        rcParams.update(rcParamsDefault)
        self.props.selected_stylename = self.get_selected_style().get_name()
        old_style = self._selected_style_params
        self._update_system_style()
        self._update_selected_style()
        data = self.props.application.get_data()
        if old_style is not None and override:
            old_colors = old_style["axes.prop_cycle"].by_key()["color"]
            color_cycle = self._selected_style_params["axes.prop_cycle"
                                                      ].by_key()["color"]
            for item_ in data:
                item_.reset(old_style, self._selected_style_params)
            count = 0
            for item_ in data:
                if isinstance(item_, item.DataItem) \
                        and item_.get_color() in old_colors:
                    if count > len(color_cycle):
                        count = 0
                    item_.set_color(color_cycle[count])
                    count += 1

        window = self.props.application.get_window()
        canvas = graphs.canvas.Canvas(self._selected_style_params)
        figure_settings = data.get_figure_settings()
        for prop in dir(figure_settings.props):
            if prop not in ("use_custom_style", "custom_style"):
                figure_settings.bind_property(prop, canvas, prop, 1 | 2)
        data.bind_property("items", canvas, "items", 2)
        from graphs.figure_settings import FigureSettingsDialog

        def on_edit_request(_canvas, label_id):
            FigureSettingsDialog(self.props.application, label_id)

        def on_view_changed(_canvas):
            data.add_view_history_state()

        canvas.connect("edit_request", on_edit_request)
        canvas.connect("view_changed", on_view_changed)

        # Set headerbar color and contrast
        bg_color = self._selected_style_params["figure.facecolor"]
        color = self._selected_style_params["text.color"]
        css_provider = window.get_headerbar_provider()
        css_provider.load_from_string(
            f"headerbar {{ background-color: {bg_color}; color: {color}; }}",
        )

        window.set_canvas(canvas)
        window.get_cut_button().bind_property(
            "sensitive",
            canvas,
            "highlight_enabled",
            2,
        )

    def _update_system_style(self) -> None:
        system_style = self._system_style_name
        if Adw.StyleManager.get_default().get_dark():
            system_style += " Dark"
        filename = _generate_filename(system_style)
        self._system_style_params = style_io.parse(
            Gio.File.new_for_uri(
                "resource:///se/sjoerd/Graphs/styles/" + filename,
            ),
        )[0]

    def _update_selected_style(self) -> None:
        self._selected_style_params = None
        if self.props.use_custom_style:
            stylename = self.props.custom_style
            for style in self.props.selection_model.get_model():
                if stylename == style.get_name():
                    try:
                        style_params = style_io.parse(style.get_file())[0]
                        if style.get_mutable():
                            style_params = self._complete_style(style_params)
                        self._selected_style_params = style_params
                        return
                    except (ValueError, SyntaxError, AttributeError):
                        self._reset_selected_style(
                            _(
                                f"Could not parse {stylename}, loading "
                                "system preferred style",
                            ).format(stylename=stylename),
                        )
                    break
            if self._selected_style_params is None:
                self._reset_selected_style(
                    _(
                        f"Plot style {stylename} does not exist "
                        "loading system preferred",
                    ).format(stylename=stylename),
                )
        self._selected_style_params = self._system_style_params

    def _reset_selected_style(self, message: str) -> None:
        self.props.use_custom_style = False
        self.props.custom_style = self._system_style_name
        self.props.application.get_window().add_toast_string(message)

    def copy_style(self, template: str, new_name: str) -> None:
        """Copy a style."""
        new_name = Graphs.tools_get_duplicate_string(
            new_name,
            self._stylenames,
        )
        destination = self.props.style_dir.get_child_for_display_name(
            _generate_filename(new_name),
        )
        for style in self.props.selection_model.get_model():
            if template == style.get_name():
                style_params = style_io.parse(style.get_file())[0]
                source = self._complete_style(style_params) \
                    if style.get_mutable() else style_params
                break
        style_io.write(destination, new_name, source)

    def _complete_style(self, params: RcParams) -> RcParams:
        for key, value in self._system_style_params.items():
            if key not in params:
                params[key] = value
        return params

    def delete_style(self, file: Gio.File) -> None:
        """Delete a style."""
        style_model = self.props.selection_model.get_model()
        for index, style in enumerate(style_model):
            if style is not None:
                file2 = style.get_file()
                if file2 is not None and file.equal(file2):
                    stylename = style.get_name()
                    style_model.remove(index)
                    self._stylenames.remove(stylename)
                    self.props.use_custom_style = False
                    self.props.custom_style = self._system_style_name


STYLE_DICT = {
    "linestyle": ["lines.linestyle"],
    "linewidth": ["lines.linewidth"],
    "markers": ["lines.marker"],
    "markersize": ["lines.markersize"],
    "draw_frame": [
        "axes.spines.bottom",
        "axes.spines.left",
        "axes.spines.top",
        "axes.spines.right",
    ],
    "tick_direction": ["xtick.direction", "ytick.direction"],
    "minor_ticks": ["xtick.minor.visible", "ytick.minor.visible"],
    "major_tick_width": ["xtick.major.width", "ytick.major.width"],
    "minor_tick_width": ["xtick.minor.width", "ytick.minor.width"],
    "major_tick_length": ["xtick.major.size", "ytick.major.size"],
    "minor_tick_length": ["xtick.minor.size", "ytick.minor.size"],
    "tick_bottom": ["xtick.bottom"],
    "tick_left": ["ytick.left"],
    "tick_top": ["xtick.top"],
    "tick_right": ["ytick.right"],
    "show_grid": ["axes.grid"],
    "grid_linewidth": ["grid.linewidth"],
    "value_padding": [
        "xtick.major.pad",
        "xtick.minor.pad",
        "ytick.major.pad",
        "ytick.minor.pad",
    ],
    "label_padding": ["axes.labelpad"],
    "title_padding": ["axes.titlepad"],
    "axis_width": ["axes.linewidth"],
    "text_color": [
        "text.color",
        "axes.labelcolor",
        "xtick.labelcolor",
        "ytick.labelcolor",
    ],
    "tick_color": ["xtick.color", "ytick.color"],
    "axis_color": ["axes.edgecolor"],
    "grid_color": ["grid.color"],
    "grid_opacity": ["grid.alpha"],
    "background_color": ["axes.facecolor"],
    "outline_color": ["figure.facecolor", "figure.edgecolor"],
}
VALUE_DICT = {
    "linestyle": ["none", "solid", "dotted", "dashed", "dashdot"],
    "markers": [
        "none",
        ".",
        ",",
        "o",
        "v",
        "^",
        "<",
        ">",
        "8",
        "s",
        "p",
        "*",
        "h",
        "H",
        "+",
        "x",
        "D",
        "d",
        "|",
        "_",
        "P",
        "X",
    ],
    "tick_direction": ["in", "out"],
}
FONT_STYLE_DICT = {
    0: "normal",
    1: "oblique",
    2: "italic",
}
FONT_VARIANT_DICT = {
    0: "normal",
    1: "small-caps",
}


def _title_format_function(_scale, value: float) -> str:
    """Format a float value as percentage string."""
    return str(value / 2 * 100).split(".")[0] + "%"


@Gtk.Template(resource_path="/se/sjoerd/Graphs/ui/style-editor.ui")
class StyleEditor(Adw.NavigationPage):
    """Style editor widget."""

    __gtype_name__ = "GraphsStyleEditor"

    style_name = Gtk.Template.Child()
    font_chooser = Gtk.Template.Child()
    titlesize = Gtk.Template.Child()
    labelsize = Gtk.Template.Child()
    linestyle = Gtk.Template.Child()
    linewidth = Gtk.Template.Child()
    markers = Gtk.Template.Child()
    markersize = Gtk.Template.Child()
    draw_frame = Gtk.Template.Child()
    tick_direction = Gtk.Template.Child()
    minor_ticks = Gtk.Template.Child()
    major_tick_width = Gtk.Template.Child()
    minor_tick_width = Gtk.Template.Child()
    major_tick_length = Gtk.Template.Child()
    minor_tick_length = Gtk.Template.Child()
    tick_bottom = Gtk.Template.Child()
    tick_left = Gtk.Template.Child()
    tick_top = Gtk.Template.Child()
    tick_right = Gtk.Template.Child()
    show_grid = Gtk.Template.Child()
    grid_linewidth = Gtk.Template.Child()
    grid_opacity = Gtk.Template.Child()
    value_padding = Gtk.Template.Child()
    label_padding = Gtk.Template.Child()
    title_padding = Gtk.Template.Child()
    axis_width = Gtk.Template.Child()
    text_color = Gtk.Template.Child()
    tick_color = Gtk.Template.Child()
    axis_color = Gtk.Template.Child()
    grid_color = Gtk.Template.Child()
    background_color = Gtk.Template.Child()
    outline_color = Gtk.Template.Child()

    line_colors_box = Gtk.Template.Child()
    poor_contrast_warning = Gtk.Template.Child()

    def __init__(self, parent):
        super().__init__()
        self.style = None
        self.parent = parent
        self._style_manager = \
            parent.get_application().get_figure_style_manager()

        self.titlesize.set_format_value_func(_title_format_function)
        self.labelsize.set_format_value_func(_title_format_function)

        # color buttons
        self.color_buttons = [
            self.text_color,
            self.tick_color,
            self.axis_color,
            self.grid_color,
            self.background_color,
            self.outline_color,
        ]
        for button in self.color_buttons:
            button.connect("clicked", self.on_color_change)
            button.provider = Gtk.CssProvider()
            button.get_style_context().add_provider(
                button.provider,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
            )

        parent.connect("load-style-request", self._load_style)
        parent.connect("save-style-request", self._save_style)

    def _load_style(self, _parent, style):
        """Load a style."""
        if not style.get_mutable():
            return
        self.style = style
        self.style_params = self._style_manager._complete_style(
            style_io.parse(self.style.get_file())[0],
        )
        stylename = self.style.get_name()
        self.set_title(stylename)
        self.style_name.set_text(stylename)
        for key, value in STYLE_DICT.items():
            value = self.style_params[value[0]]
            with contextlib.suppress(KeyError):
                value = VALUE_DICT[key].index(value)
            widget = getattr(self, key.replace("-", "_"))
            if isinstance(widget, Adw.EntryRow):
                widget.set_text(str(value))
            elif isinstance(widget, Adw.ComboRow):
                widget.set_selected(int(value))
            elif isinstance(widget, Gtk.Scale):
                widget.set_value(value)
            elif isinstance(widget, Gtk.Button):
                widget.color = Graphs.tools_hex_to_rgba(value)
            elif isinstance(widget, Adw.SwitchRow):
                widget.set_active(bool(value))
            else:
                raise ValueError

        # font
        font_description = Pango.FontDescription.new()
        font_size = self.style_params["font.size"]
        font_description.set_size(font_size * Pango.SCALE)
        self.titlesize.set_value(
            round(self.style_params["figure.titlesize"] * 2 / font_size, 1),
        )
        self.labelsize.set_value(
            round(self.style_params["axes.labelsize"] * 2 / font_size, 1),
        )
        font_description.set_family(self.style_params["font.sans-serif"][0])
        font_description.set_weight(self.style_params["font.weight"])
        inverted_style_dict = {b: a for a, b in FONT_STYLE_DICT.items()}
        font_description.set_style(
            inverted_style_dict[self.style_params["font.style"]],
        )
        inverted_variant_dict = {b: a for a, b in FONT_VARIANT_DICT.items()}
        font_description.set_variant(
            inverted_variant_dict[self.style_params["font.variant"]],
        )
        self.font_chooser.set_font_desc(font_description)

        for button in self.color_buttons:
            hex_color = Graphs.tools_rgba_to_hex(button.color)
            button.provider.load_from_string(
                f"button {{ color: {hex_color}; }}",
            )
        self._check_contrast()

        # line colors
        self.line_colors = \
            self.style_params["axes.prop_cycle"].by_key()["color"]
        self._reload_line_colors()

    def _save_style(self, _parent):
        """Save the style."""
        if self.style is None:
            return
        for key in STYLE_DICT.keys():
            widget = getattr(self, key.replace("-", "_"))
            if isinstance(widget, Adw.EntryRow):
                value = str(widget.get_text())
            elif isinstance(widget, Adw.ComboRow):
                value = widget.get_selected()
            elif isinstance(widget, Gtk.Scale):
                value = widget.get_value()
            elif isinstance(widget, Gtk.Button):
                value = Graphs.tools_rgba_to_hex(widget.color)
            elif isinstance(widget, Adw.SwitchRow):
                value = bool(widget.get_active())
            else:
                raise ValueError
            with contextlib.suppress(KeyError):
                value = VALUE_DICT[key][value]
            for item_ in STYLE_DICT[key]:
                self.style_params[item_] = value

        # font
        font_description = self.font_chooser.get_font_desc()
        self.style_params["font.sans-serif"] = [font_description.get_family()]
        font_size = font_description.get_size() / Pango.SCALE
        for key in (
            "font.size",
            "xtick.labelsize",
            "ytick.labelsize",
            "legend.fontsize",
            "figure.labelsize",
        ):
            self.style_params[key] = font_size
        titlesize = round(self.titlesize.get_value() / 2 * font_size, 1)
        labelsize = round(self.labelsize.get_value() / 2 * font_size, 1)
        self.style_params["figure.titlesize"] = titlesize
        self.style_params["axes.titlesize"] = titlesize
        self.style_params["axes.labelsize"] = labelsize
        font_weight = font_description.get_weight()
        for key in (
            "font.weight",
            "axes.titleweight",
            "axes.labelweight",
            "figure.titleweight",
            "figure.labelweight",
        ):
            self.style_params[key] = font_weight
        self.style_params["font.style"] = \
            FONT_STYLE_DICT[font_description.get_style()]
        self.style_params["font.variant"] = \
            FONT_VARIANT_DICT[font_description.get_variant()]

        # line colors
        self.style_params["axes.prop_cycle"] = cycler(color=self.line_colors)
        self.style_params["patch.facecolor"] = self.line_colors[0]

        # name & save
        new_name = self.style_name.get_text()
        if self.style.get_name() != new_name:
            new_name = Graphs.tools_get_duplicate_string(
                new_name,
                self._style_manager.list_stylenames(),
            )
        style_io.write(self.style.get_file(), new_name, self.style_params)
        self._style_manager._on_style_change(True)
        self.style = None

    def _reload_line_colors(self):
        list_box = self.line_colors_box
        while list_box.get_last_child() is not None:
            list_box.remove(list_box.get_last_child())
        if self.line_colors:
            for index in range(len(self.line_colors)):
                list_box.append(_StyleColorBox(self, index))
        else:
            self.line_colors.append("#000000")
            list_box.append(_StyleColorBox(self, 0))

    def on_color_change(self, button):
        """Handle color change."""

        def on_accept(dialog, result):
            with contextlib.suppress(GLib.GError):
                color = dialog.choose_rgba_finish(result)
                if color is not None:
                    button.color = color
                    self._check_contrast()
                    hex_color = Graphs.tools_rgba_to_hex(button.color)
                    button.provider.load_from_string(
                        f"button {{ color: {hex_color}; }}",
                    )

        dialog = Gtk.ColorDialog()
        dialog.set_with_alpha(False)
        dialog.choose_rgba(
            self.parent.get_application().get_window(),
            button.color,
            None,
            on_accept,
        )

    def _check_contrast(self):
        contrast = Graphs.tools_get_contrast(
            self.outline_color.color,
            self.text_color.color,
        )
        self.poor_contrast_warning.set_visible(contrast < 4.5)

    @Gtk.Template.Callback()
    def on_linestyle(self, comborow, _b):
        """Handle linestyle selection."""
        self.linewidth.set_sensitive(comborow.get_selected() != 0)

    @Gtk.Template.Callback()
    def on_markers(self, comborow, _b):
        """Handle marker selection."""
        self.markersize.set_sensitive(comborow.get_selected() != 0)

    @Gtk.Template.Callback()
    def add_color(self, _button):
        """Add a color."""
        self.line_colors.append("000000")
        self._reload_line_colors()

    @Gtk.Template.Callback()
    def on_delete(self, _button):
        """Handle style deletion."""

        def on_response(_dialog, response):
            if response == "cancel_delete_style":
                return
            if response == "delete_style":
                file = self.style.get_file()
                self._style_manager.delete_style(file)
                file.trash(None)
                self.style = None
                self.parent.get_navigation_view().pop()

        dialog = Graphs.tools_build_dialog("delete_style_dialog")
        msg = _("Are you sure you want to delete {stylename}?")
        dialog.set_body(msg.format(stylename=self.style.get_name()))
        dialog.connect("response", on_response)
        dialog.present(self.parent)


@Gtk.Template(resource_path="/se/sjoerd/Graphs/ui/style-color-box.ui")
class _StyleColorBox(Gtk.Box):
    __gtype_name__ = "GraphsStyleColorBox"
    label = Gtk.Template.Child()
    color_button = Gtk.Template.Child()

    parent = GObject.Property(type=StyleEditor)
    index = GObject.Property(type=int, default=0)

    def __init__(self, parent, index):
        super().__init__(parent=parent, index=index)
        self.label.set_label(_("Color {number}").format(number=index + 1))
        self.provider = Gtk.CssProvider()
        self.color_button.get_style_context().add_provider(
            self.provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )
        self._reload_color()

    def _reload_color(self):
        color = self.props.parent.line_colors[self.props.index]
        self.provider.load_from_string(f"button {{ color: {color}; }}")

    @Gtk.Template.Callback()
    def on_color_choose(self, _button):

        def on_accept(dialog, result):
            with contextlib.suppress(GLib.GError):
                color = dialog.choose_rgba_finish(result)
                if color is not None:
                    self.props.parent.line_colors[self.props.index] = \
                        Graphs.tools_rgba_to_hex(color)
                    self._reload_color()

        dialog = Gtk.ColorDialog()
        dialog.set_with_alpha(False)
        dialog.choose_rgba(
            self.props.parent.parent,
            Graphs.tools_hex_to_rgba(
                self.props.parent.line_colors[self.props.index],
            ),
            None,
            on_accept,
        )

    @Gtk.Template.Callback()
    def on_delete(self, _button):
        self.props.parent.line_colors.pop(self.props.index)
        self.props.parent._reload_line_colors()



# SPDX-License-Identifier: GPL-3.0-or-later
"""Generate Graphs GResource at build time."""
import argparse
import importlib.util
import logging
import shutil
import sys
from pathlib import Path
from xml.etree import ElementTree

from PIL import Image

from gi.repository import Gio

from matplotlib import font_manager

import numpy

parser = argparse.ArgumentParser(
    description="Generate Graphs GResource at build time.",
)
parser.add_argument(
    "out",
    help="The output file",
)
parser.add_argument(
    "dir",
    help="Path to build directory",
)
parser.add_argument(
    "style_io",
    help="Path to style_io.py",
)
parser.add_argument(
    "--ui",
    required=True,
    nargs="+",
    dest="ui",
    help="List of UI Files.",
)
parser.add_argument(
    "--styles",
    required=True,
    nargs="+",
    dest="styles",
    help="List of Styles.",
)
parser.add_argument(
    "--other",
    required=True,
    nargs="+",
    dest="other",
    help="List of other Files to include at toplevel.",
)
parser.add_argument(
    "--icons",
    required=True,
    nargs="+",
    dest="icons",
    help="List of icon Files.",
)
args = parser.parse_args()

# Check fonts
font_list = font_manager.findSystemFonts(fontpaths=None, fontext="ttf")
for font in font_list:
    try:
        font_manager.fontManager.addfont(font)
    except RuntimeError:
        logging.warning("Could not load %s", font)
# Disable matplotlib logging
logging.getLogger("matplotlib.font_manager").setLevel(logging.ERROR)

# dynamically import style_io
spec = importlib.util.spec_from_file_location("style_io", args.style_io)
style_io = importlib.util.module_from_spec(spec)
sys.modules["style_io"] = style_io
spec.loader.exec_module(style_io)

# GResource tree creation
gresources = ElementTree.Element("gresources")
main_gresource = ElementTree.SubElement(
    gresources,
    "gresource",
    attrib={"prefix": "/se/sjoerd/Graphs/"},
)

# Begin Other Section
for file in args.other:
    name = Path(shutil.copy(file, args.dir)).name
    element = ElementTree.SubElement(
        main_gresource,
        "file",
        attrib={
            "compressed": "True",
        },
    )
    element.text = name
# End Other Section

# Begin style section
styles = {}
styles_gresource = ElementTree.SubElement(
    gresources,
    "gresource",
    attrib={"prefix": "/se/sjoerd/Graphs/styles/"},
)
for style_path in args.styles:
    style_file = shutil.copy(style_path, args.dir)
    name = Path(style_file).name
    style_element = ElementTree.SubElement(
        styles_gresource,
        "file",
        attrib={
            "compressed": "True",
        },
    )
    style_element.text = name
    params_file = Gio.File.new_for_path(style_file)
    params, stylename = style_io.parse(params_file)
    out_path = Path(args.dir, name.replace(".mplstyle", ".png"))
    styles[stylename] = out_path
    with open(out_path, "wb") as out_file:
        style_io.create_preview(out_file, params, "png")
    preview_element = ElementTree.SubElement(
        main_gresource,
        "file",
        attrib={
            "compressed": "True",
        },
    )
    preview_element.text = out_path.name


def _to_array(file_path):
    with open(file_path, "rb") as file:
        return numpy.array(Image.open(file).convert("RGB"))


# Generate stitched system previews for Adwaita and Yaru
for sys_style in ("Adwaita", "Yaru"):
    light_array = _to_array(styles[sys_style])
    dark_array = _to_array(styles[sys_style + " Dark"])
    height, width = light_array.shape[0:2]
    stitched_array = numpy.concatenate(
        (light_array[:, :width // 2], dark_array[:, width // 2:]),
        axis=1,
    )
    stitched_image = Image.fromarray(stitched_array)
    out_path = Path(args.dir + "/system-style-" + sys_style.lower() + ".png")
    with open(out_path, "wb") as file:
        stitched_image.save(file, "PNG")
    preview_element = ElementTree.SubElement(
        main_gresource,
        "file",
        attrib={
            "compressed": "True",
        },
    )
    preview_element.text = out_path.name
# End style section

# Begin ui section
ui_gresource = ElementTree.SubElement(
    gresources,
    "gresource",
    attrib={"prefix": "/se/sjoerd/Graphs/ui/"},
)
help_overlay_path = None
for ui_file in args.ui:
    path = Path(ui_file)
    if path.name == "shortcuts.ui":
        help_overlay_path = path
        continue
    ui_file_element = ElementTree.SubElement(
        ui_gresource,
        "file",
        attrib={
            "preprocess": "xml-stripblanks",
            "alias": path.name,
        },
    )
    ui_file_element.text = "ui/" + path.name
help_overlay_element = ElementTree.SubElement(
    main_gresource,
    "file",
    attrib={
        "preprocess": "xml-stripblanks",
        "alias": "gtk/help-overlay.ui",
    },
)
help_overlay_element.text = "ui/" + help_overlay_path.name
# End ui section

# Begin icon section
icon_gresource = ElementTree.SubElement(
    gresources,
    "gresource",
    attrib={"prefix": "/se/sjoerd/Graphs/icons/scalable/actions/"},
)
for icon_file in args.icons:
    path = Path(Path(shutil.copy(icon_file, args.dir)))
    icon_file_element = ElementTree.SubElement(
        icon_gresource,
        "file",
        attrib={
            "preprocess": "xml-stripblanks",
        },
    )
    icon_file_element.text = path.name
# End icon section

# Write
tree = ElementTree.ElementTree(gresources)
tree.write(args.out)



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for data Items."""
from gi.repository import GObject, Graphs

from graphs import misc


def new_from_dict(dictionary: dict):
    """Instanciate item from dict."""
    match dictionary["type"]:
        case "GraphsDataItem":
            cls = DataItem
        case "GrapsTextItem":
            cls = TextItem
        case "GrapsFillItem":
            cls = FillItem
        case _:
            pass
    dictionary.pop("type")
    return cls(**dictionary)


class _ItemMixin():

    def reset(self, old_style, new_style):
        """Reset all properties."""
        for prop, (key, function) in self._style_properties.items():
            old_value = old_style[key]
            new_value = new_style[key]
            if function is not None:
                old_value = function(old_value)
                new_value = function(new_value)
            if self.get_property(prop) == old_value:
                self.set_property(prop, new_value)

    def _extract_params(self, style):
        return {
            prop: style[key] if function is None else function(style[key])
            for prop, (key, function) in self._style_properties.items()
        }

    def to_dict(self):
        """Convert item to dict."""
        dictionary = {key: self.get_property(key) for key in dir(self.props)}
        dictionary["type"] = self.__gtype_name__
        return dictionary


class DataItem(Graphs.Item, _ItemMixin):
    """DataItem."""

    __gtype_name__ = "GraphsDataItem"

    xdata = GObject.Property(type=object)
    ydata = GObject.Property(type=object)
    linestyle = GObject.Property(type=int, default=1)
    linewidth = GObject.Property(type=float, default=3)
    markerstyle = GObject.Property(type=int, default=0)
    markersize = GObject.Property(type=float, default=7)

    _style_properties = {
        "linestyle": ("lines.linestyle", misc.LINESTYLES.index),
        "linewidth": ("lines.linewidth", None),
        "markerstyle": ("lines.marker", misc.MARKERSTYLES.index),
        "markersize": ("lines.markersize", None),
    }

    @classmethod
    def new(cls, style, xdata=None, ydata=None, **kwargs):
        """Create new DataItem."""
        return cls(
            xdata=xdata,
            ydata=ydata,
            **cls._extract_params(cls, style),
            **kwargs,
        )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        for prop in ("xdata", "ydata"):
            if self.get_property(prop) is None:
                self.set_property(prop, [])


class TextItem(Graphs.Item, _ItemMixin):
    """TextItem."""

    __gtype_name__ = "GraphsTextItem"

    xanchor = GObject.Property(type=float, default=0)
    yanchor = GObject.Property(type=float, default=0)
    text = GObject.Property(type=str, default="")
    size = GObject.Property(type=float, default=12)
    rotation = GObject.Property(type=int, default=0, minimum=0, maximum=360)

    _style_properties = {
        "size": ("font.size", None),
        "color": ("text.color", None),
    }

    @classmethod
    def new(cls, style, xanchor=0, yanchor=0, text="", **kwargs):
        """Create new textItem."""
        return cls(
            xanchor=xanchor,
            yanchor=yanchor,
            text=text,
            **cls._extract_params(cls, style),
            **kwargs,
        )


class FillItem(Graphs.Item, _ItemMixin):
    """FillItem."""

    __gtype_name__ = "GraphsFillItem"

    data = GObject.Property(type=object)

    @classmethod
    def new(cls, _params, data, **kwargs):
        """Create new FillItem."""
        return cls(data=data, **kwargs)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.props.data is None:
            self.props.data = (None, None, None)

    def reset(self):
        """Not yet implemented."""
        raise NotImplementedError
